#ifndef _VALUES_COMMON_H_
#define _VALUES_COMMON_H_

#include <velib/types/ve_item.h>
#include <velib/types/ve_item_def.h>
#include <velib/types/variant_print.h>
#include <msg_router.h>

struct VeSystemS;
struct VeSettingS;

typedef void VeSystemCall(struct VeSystemS *veSystem);

typedef struct {
	VeItem productId;
	VeItem *ptrProductName;
	VeItem productName;
	VeItem firmwareVersion;
	VeItem hardwareVersion;
	VeItem instance;
	VeItem connected;
	VeItem serial;
	VeItem n2kUnique;
} VeProductInfo;

/* This context is passed to the setting set function */
typedef struct
{
	struct VeSystemS *system;
	struct VeSettingS *settings;
} VeSettingCtx;

/**
 * Some victron devices can be grouped to make up a system. For example
 * mutiple Multi's can be put in parallel to increase the total power.
 * Also Skylla-I chargers / Solar chargers can be but in parrallel and
 * will act as a single bigger charger. Commonly the total what such a
 * group does is of more interest than what the inidividual device do,
 * e.g. total charge current, solar yield etc is more informative then
 * reporting all per device values.
 */
typedef struct VeSystemS
{
	/* Pointer to the next system */
	struct VeSystemS *next;
	/* Sub-system(s) */
	struct VeSystemS *child;
	struct VeSystemS *parent;
	/* The master device */
	struct J1939Func *master;
	/* The system / product dependent CAN bus variables */
	void *canbusCtx;
	/* Settings context */
	VeSettingCtx settingsCtx;
	/*
	 * This is is the root of the item tree. For example a voltage
	 * from this system can be lookep up as /Dc/V, see ve_item for
	 * details.
	 */
	VeItem values;

	/* Non-product related items */
	VeItem processName;
	VeItem processVersion;
	VeItem connection;

	/* Product related items (master) */
	VeProductInfo product;

	/*
	 * The index of above tree in the global tree, e.g. com.victronegy.bmv.
	 * This allows to have access to all devices by a single interface. For
	 * example if the above voltage can be looked up with
	 * com.victronenergy.bmv/Dc/V
	 */
	char const *id;

	N2kFunctionElectrical function;
	un8 group;
	un8 instance;

	VeSystemCall *init;
	VeSystemCall *update;
	VeSystemCall *tick;	/* timer callback 50ms */
} VeSystem;

typedef struct
{
	float min;
	float max;
	float def;
} VeRange;

typedef struct VeSettingS
{
	VeRegId id;
	int offset;
	VeDataBasicType type;
	VeRange const *range;
	VeScale const *scale;

	veBool (*isLocked)(void *ctx);
} VeSetting;

/**
 * To seperate gui logic / any other logic from the communication
 * the received values are stored in a tree in SI units. This struct
 * contains the field needed to create the tree.
 */
typedef struct
{
	char const *id;
	VeVariantUnitFmt *fmt;
	int offset;
	un8 timeout;
} ItemInfo;

typedef struct
{
	VeItem voltage;
	VeItem current;
	VeItem power;
} DcStatus;

/* helper to create item trees */
void valuesInitTree(ItemInfo const *itemInfo, size_t count, VeItem *root, VeSystem *system);
VeSetting *valuesGetSettingByItem(VeSettingCtx *ctx, VeItem *needle);

VE_DCL void valuesInit(VeItem **mainNad);
VE_DCL void valuesUpdate(void);
VE_DCL void valuesTick(void);

#endif
