#include <velib/velib_inc_J1939.h>

#include <velib/canhw/canhw.h>
#include <velib/io/ve_stream.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/io/ve_remote_device_vecan.h>
#include <velib/J1939/J1939.h>
#include <velib/J1939/J1939_device.h>
#include <velib/J1939/J1939_function.h>
#include <velib/J1939/J1939_nmt.h>
#include <velib/J1939/J1939_request.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_fp.h>
#include <velib/platform/task.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_todo.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/prod_data.h>
#include <velib/vecan/prod_data_cm.h>
#include <velib/vecan/vecan.h>
#include <velib/vecan/reg_rx.h>

#include <msg_router.h>
#include <senders.h>
#include <vecan_common.h>

/* A hook can be register to receive messages from a single device */
static MsgRouterHook *hooks;
extern J1939DeviceTp *sendingDevices;

static veBool hookExists(MsgRouterHook *hook)
{
	MsgRouterHook *h = hooks;

	while (h) {
		if (h == hook)
			return veTrue;
		h = h->next;
	}

	return veFalse;
}

void msgRouterAddHook(MsgRouterHook *hook)
{
	if (!hookExists(hook)) {
		hook->next = hooks;
		hooks = hook;
	}
}

void msgRouterRemoveHook(MsgRouterHook *hook)
{
	MsgRouterHook *h = hooks;

	if (hooks == hook) {
		hooks = hooks->next;
		hook->next = NULL;
		return;
	}

	while (h->next) {
		if (h->next == hook) {
			h->next = hook->next;
			hook->next = NULL;
			return;
		}
		h = h->next;
	}
}

static void veRegHandler(VeEvent ev)
{
	MsgRouterHook* hook = hooks;

	while (hook) {
		if (hook->func && hook->func->nmt.nad == n2kMsgIn->msgId.p.src &&
				(hook->func->nmt.flags & J1939_NMT_DEV_DISCONNECTED) == 0 && hook->vregFun)
			hook->vregFun(hook->ctx, hook->func, ev, veRegRxId, veRegAckCode);
		hook = hook->next;
	}
}

/* Vreg handlers */
VeEventHndlr const veRegRxHndlrDef[] =
{
	veRegHandler,
	veDevUdfDataVeRegs,
	veCanRemoteDevRegHandler,
	veMsgHandlerVeRegs, /* keep as last */
	NULL
};

/** Attempt to push a queued message to the CANhw (if any). */
void msgRouterUpdateTx(void)
{
	J1939Msg* msg;

	if ((msg = j1939TxPop()) != NULL)
	{
		veBool status;
		// try to send the message with the fastpacket protocl
		if (msg->type.p.kind == N2K_FP)
			status = n2kFpSend( (N2kFp*) msg );
		// just send the message
		else
			status = j1939SfSend( (J1939Sf*) msg );

		// if the message is send, free it again
		if (status)
		{
			veTodo();
			j1939TxFree(msg);
		} else {
			j1939TxPushFront(msg);
		}
	}
}

/**
 * Handles J1939-requests as the input messages.
 *
 * Should only be called after it has been verfied that the incoming
 * message is J1939-request. If the request cannot be fullfilled a NACK
 * will be queued for output when the request was addressed.
 */
static void j1939Request(void)
{
	J1939Pgn pgn;

	if (n2kMsgIn->length < 3)
		return;

	// get the pgn being requested
	pgn = J1939PgnPfPsNoTg(n2kMsgIn->mdata[1], n2kMsgIn->mdata[0]);

	// handle the message
	if (n2kMsgIn->mdata[2] == 0)
	{
		switch (pgn)
		{
		// our ACL is requested so send it
		case J1939_PGN0_ACL:
			j1939AclUpdateState(J1939_ACL_EV_REQ);
			return;

		default:
			if (veProdDataMarkPgn(0, pgn))
				return;
			break;

		}
	} else if (n2kMsgIn->mdata[2] == 1) {
		switch (pgn)
		{
		default:
			if (veProdDataMarkPgn(1, pgn))
				return;
			break;
		}

	}

	if (n2kMsgIn->msgId.p.ps == J1939_NAD_BROADCAST)
		return;

	j1939SendNack(n2kMsgIn->msgId.p.src, n2kMsgIn->mdata[2], pgn);
}


/**
 * Forwards any input message on n2kMsgIn.
 *
 * The message can be any kind of message. An incoming message or a
 * reconstructed or a loopback. The message is forwarded based on its
 * CAN identifier. This handles the message for the main-device, so
 * either this is done ones per message. Per device message are handled
 * below, see handleRxSendingDevices.
 */
static void handleRxMainDevice(void)
{
	J1939Id canId;

	// by default message are only processed once and then freed.
	canId = n2kMsgIn->msgId;

	// Remove target address for device specific messages.
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
	{
		// handle incoming message for the data page 0.
		switch(J1939PgnFromId(canId))
		{
		// another device send a ACL
		case J1939_PGN0_ACL:
			j1939NmtData();
			break;

		case PGN0_VECAN_VREG:
			if (veInUn16Le() != N2K_MAN_TAG_VICTRON)
				return;
			veRegRxData();
			break;

		default:
			;
		}
	} else {
		/* handle incoming message for data page 1 */
		switch (J1939PgnFromId(canId))
		{
		case N2K_PGN1_PRODUCT_INFO:
			veDevN2kDataProdInfo();
			break;

		case PGN0_VECAN_VREG:
			if (veInUn16Le() != N2K_MAN_TAG_VICTRON)
				return;
			veRegRxData();
			break;

		default:
			;
		}
	}
}

/**
 * @brief handleRxSendingDevices
 *
 * Handles incoming requests / data for the CAN-bus producers (iow sending
 * runtime data towards the CAN-bus.
 */
static void handleRxSendingDevices(void)
{
	J1939DeviceTp *dev;
	J1939Id canId;

	// by default message are only processed once and then freed.
	canId = n2kMsgIn->msgId;

	// Remove target address for device specific messages.
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	/* Update the device claimed here */
	dev = sendingDevices;
	while (dev) {
		j1939DevActivate(dev);

		if (j1939MsgForDevice(&n2kMsgIn->msgId))
		{
			veInMove(0);

			if (n2kMsgIn->msgId.p.pg.p.dp == 0)
			{
				switch (J1939PgnFromId(canId))
				{
				// some one send a request to this device.
				case J1939_PGN0_REQUEST:
					j1939Request();
					break;

				case J1939_PGN0_ACL:
					j1939AclData();
					if (dev->flags & J1939_ACL_STORE_NAD) {
						VeVariant v;
						veItemSet(dev->nadItem, veVariantUn8(&v, dev->nad));
						dev->flags &= ~J1939_ACL_STORE_NAD;
					}
					break;

				default:
					;
				}
			} else {
				/* handle incoming message for data page 1 */
				switch (J1939PgnFromId(canId))
				{
				case N2K_PGN1_COMPLEX_GROUP:
					veProdDataMsg();
					break;

				default:
					;
				}
			}
		}
		dev = dev->next;
	}
}

static void handleRx(void)
{
	handleRxSendingDevices();

	mainDeviceActivate();

	if (j1939MsgForDevice(&n2kMsgIn->msgId)) {
		veInMove(0);
		handleRxMainDevice();
	}
}

/**
 * Process queued rx messages from the CANhw till none is left.
 */
void msgRouterUpdateRx(void)
{
	J1939Sf sf;
	N2kFp* n2kFp;

	sf.type.w = J1939_SF;

	for (;;)
	{
		MsgRouterHook* hook = hooks;
		J1939FuncTp* dev;

		if (!veCanRead((VeRawCanMsg*) &sf))
			break;

		j1939AclDetectConflict(&sf.msgId);

		/* warning dev might be NULL! */
		dev = j1939FuncByNad(sf.msgId.p.src);

		n2kFp = NULL;
		if (sf.type.p.kind == J1939_SF && n2kIsFp(&sf.msgId))
		{
			/*
			 * Reconstruct fast packets... (allows sending longer messages)
			 * If a message is reconstructed, it is processed next.
			 */
			if ( (n2kFp = n2kFpData( (J1939Sf*) &sf )) != NULL) {
				veInMsg((J1939Msg*) n2kFp);
				j1939NmtActivityDev(dev);
				handleRx();
			} else {
				/* this message did not complete a fastpacket */
				continue;
			}
		}
		else
		{
			veInMsg((J1939Msg*) &sf);
			j1939NmtActivityDev(dev);
			handleRx();
		}

		if (dev != NULL) {
			if (dev->nmt.flags & J1939_NMT_DEV_INSTANCE_CHANGED) {
				msgRouterInstanceChanged(dev);
				dev->nmt.flags &= ~J1939_NMT_DEV_INSTANCE_CHANGED;
			}

			while (hook) {
				if (hook->rxFun && (hook->func == NULL || hook->func == dev))
				{
					//logI("", "%p", hook->func);
					veInMove(0);
					hook->rxFun(hook->ctx, dev);
				}
				hook = hook->next;
			}
		}

		if (n2kFp != NULL)
			n2kFpRxFree(n2kFp);
	}
}

void msgRouterUpdate(void)
{
	J1939FuncTp* veDev = j1939FuncList;
	J1939DeviceTp *dev;

	/* Handle incoming frames */
	msgRouterUpdateRx();
	/* Try to send responses asap */
	msgRouterUpdateTx();

	j1939NmtUpdate();

	/* Update the device claimed here */
	dev = sendingDevices;
	while (dev) {
		j1939DevActivate(dev);
		j1939AclUpdate();
		veProdDataUpdate();
		dev = dev->next;
	}
	mainDeviceActivate();

	/* update data receivers */
	while (veDev) {
		veDevUdfUpdate(veDev);
		veDevN2kUpdate(veDev);
		veDev = veDev->next;
	}
}

/** 50 ms period updates. */
void msgRouterTick(void)
{
	J1939DeviceTp *dev = sendingDevices;

	n2kFpTick();
	j1939NmtTick();
	veProdDataTick();

	while (dev) {
		j1939DevActivate(dev);
		j1939AclTick();
		dev = dev->next;
	}
	mainDeviceActivate();
}

/**
 * A CAN bus message was received.
 */
veBool veCanRxEvent(void)
{
	taskUpdate();
	return veTrue;
}

void msgRouterInit(VeItem *mainNad)
{
	j1939SfInit();
	n2kFpInit();

	mainDeviceAdd(0, mainNad);
	mainDeviceActivate();
}

/**
 * Allocates a message used for transmission.
 *
 * @context main
 */
J1939Msg* j1939TxAlloc(un8 type)
{
	// start output
	switch (type)
	{
	case J1939_SF:
		return (J1939Msg*) j1939SfTxAlloc();

	case N2K_FP:
		return (J1939Msg*)  n2kFpTxAlloc();
	}

	veAssert(veFalse);
	return NULL;
}

/**
 * Frees a message used for transmission.
 *
 * @note
 * 	Fast Packet response might use the RX queue since it contains
 * 	more buffers and hence use resources a bit better.
 *
 * @context main
 */
void j1939TxFree(J1939Msg* msg)
{
	// start output
	switch (msg->type.p.kind)
	{
	case J1939_SF:
		j1939SfTxFree((J1939Sf*) msg);
		break;

	case N2K_FP:
		n2kFpTxFree((N2kFp *) msg);
		break;

	default:
		veAssert(veFalse);
		break;
	}
}
