#include <velib/velib_inc_J1939.h>

#include <velib/J1939/J1939_device.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_fp.h>
#include <velib/vecan/prod_types.h>

#include <senders.h>

#define sptr(a)		(void*)(&(a)),sizeof(a)
#define ptr(a)		(void*)(&(a))
#define fptr(a)		(void*)(&(a)),0
#define strptr(a)	(void*)(&(a[0])),sizeof(a)
#define nodata		NULL, 0

static J1939DeviceTp device;

typedef enum
{
	MAP_UNITARY,
} veScalesNames;

// NMEA200
static un16	const n2kDb				= 1301;								/* NMEA 2000 db version */
static un16	const n2kProductID		= 0xFFFF;							/* Product id */
static char	const n2kModel[]		= "VE.Can interface";				/* Model */
static char	const n2kSwVersion[]	= "1.0";							/* Software version */
static char	const n2kModelVersion[] = "1.0";							/* Model Version */
static un8	const n2kCert			= 1;								/* Certification level */
static un8	const n2kLoad			= 0;								/* Load equivalant (50mA) */
static un8 const pgnTxOpcode = 0;
static un8 const pgnRxOpcode = 1;

#define PGN(a)	(a&0xFF),(un8)(a>>8)

static un8 const pgnsRx[] =
{
	// data page 0, addressable
	PGN(J1939_PGN0_REQUEST),		0,
	PGN(J1939_PGN0_ACL),			0,
	PGN(N2K_PGN1_PROPRIETARY_PDU1),	0,

	// data page 1, addressable
	PGN(N2K_PGN1_COMPLEX_GROUP),	1,
};

static un8 const pgnsTx[] =
{
	// data page 0, addressable
	PGN(J1939_PGN0_ACK), 			0,
	PGN(J1939_PGN0_ACL), 			0,
	PGN(N2K_PGN1_PROPRIETARY_PDU1),	0,

	// data page 1, addressable
	PGN(N2K_PGN1_COMPLEX_GROUP), 	1,
	PGN(N2K_PGN1_PGN_LIST), 		1,

	// data page 1, broadcast
	PGN(N2K_PGN1_PRODUCT_INFO), 	1,
};

static VeProdSgnDef const veProdSgnDef[] =
{
#define F0_N2K_PRODUCT_INFO_S			0
#define F0_N2K_PRODUCT_INFO_C			8
#define F0_N2K_PRODUCT_INFO_E			(F0_N2K_PRODUCT_INFO_S + F0_N2K_PRODUCT_INFO_C)

	{VE_UN16,		MAP_UNITARY,			0,						sptr(n2kDb)									},
	{VE_UN16,		MAP_UNITARY,			0,						sptr(n2kProductID)							},
	{VE_CHAR32,		MAP_UNITARY,			0,						strptr(n2kModel)							},
	{VE_CHAR32,		MAP_UNITARY,			0,						strptr(n2kSwVersion)						},
	{VE_CHAR32,		MAP_UNITARY,			0,						strptr(n2kModelVersion)						},
	{VE_CHAR32,		MAP_UNITARY,			0,						strptr("serial#")							},
	{VE_UN8,		MAP_UNITARY,			0,						sptr(n2kCert)								},
	{VE_UN8,		MAP_UNITARY,			0,						sptr(n2kLoad)								},

#define N2K_PGNS_TX_S				F0_N2K_PRODUCT_INFO_E
#define N2K_PGNS_TX_C				2
#define N2K_PGNS_TX_E				(N2K_PGNS_TX_S + N2K_PGNS_TX_C)
	{VE_UN8,		MAP_UNITARY,			S_INST,					sptr(pgnTxOpcode)							},
	{VE_BUF,		MAP_UNITARY,			0,						sptr(pgnsTx)								},

#define N2K_PGNS_RX_S				N2K_PGNS_TX_E
#define N2K_PGNS_RX_C				2
#define F0_N2K_PGNS_RX_E				(N2K_PGNS_RX_S + N2K_PGNS_RX_C)
	{VE_UN8,		MAP_UNITARY,			S_INST,					sptr(pgnRxOpcode)							},
	{VE_BUF,		MAP_UNITARY,			0,						sptr(pgnsRx)								},
};


static VeProdMsgDefConst const veProdMsgDefConst[] =
{
	{N2K_FP,	1,	N2K_PGN1_PRODUCT_INFO,		F0_N2K_PRODUCT_INFO_S,		F0_N2K_PRODUCT_INFO_C},
	{N2K_FP,	1,	N2K_PGN1_PGN_LIST,			N2K_PGNS_TX_S,				N2K_PGNS_TX_C},
	{N2K_FP,	1,	N2K_PGN1_PGN_LIST,			N2K_PGNS_RX_S,				N2K_PGNS_RX_C},
};

static VeProdMsgDefInvolatile msgSettings[ARRAY_LENGTH(veProdMsgDefConst)] = {
	{0,		6}
};

static VeProdMsgDefVolatile msgMemory[ARRAY_LENGTH(veProdMsgDefConst)];

static un8 fpCounters[ARRAY_LENGTH(veProdMsgDefConst)];


/** Announce the device on the dbus after all settings are valid. */
static void announceDeviceWhenComplete(J1939DeviceTp *dev)
{
	if (dev->isAnnounced)
		return;

	if (veItemIsValid(dev->nadItem)) {
		dev->isAnnounced = veTrue;
		j1939AclEnable();
	}
}

static void nadSettingChanged(VeItem *item)
{
	VeVariant v;
	J1939DeviceTp *dev = (J1939DeviceTp *) veItemCtx(item)->ptr;
	un8 nad = veItemValue(item, &v)->value.UN8;

	j1939DevActivate(dev);
	if (!dev->isAnnounced)
	{
		dev->nad = nad;
		announceDeviceWhenComplete(dev);
	} else {
		if (nad != dev->nad)
			j1939AclChangeNad(nad);
	}

	mainDeviceActivate();
}

/**
 * The main interface, which claims an address on the bus.
 * Most code, at very least the listening code, only use this device.
 *
 * @param nad The network adddress to use initially. Which can be set later on,
 *        e.g. when received from localsettings. It will be updated if the nad
 *        on the network changed (and hence stored in case localsettings is used).
 *
 * @note The main device is currently anounced as display. This might not always be,
 *       but shouldn't matter that much..
 */
void mainDeviceAdd(un32 unique, VeItem *nad)
{
	/* normally something valid goes here .. */
	device.name.low = J1939_ACL_L_MAN(N2K_MAN_VICTRON) | J1939_ACL_L_IDENTITY_NUMBER(unique);
	device.name.high.vun32 = J1939_ACL_H_INDUSTRY(J1939_INDUSTRY_MARINE) | J1939_ACL_H_SCA |
					J1939_ACL_H_CLASS(N2K_CLASS_DISPLAY) | J1939_ACL_H_FUNCTION(N2K_FUNCTION_DISPLAY_UNIT);

	j1939DevActivate(&device);

	j1939AclInit();

	device.veProdMsgDefConst = veProdMsgDefConst;
	device.msgCount = ARRAY_LENGTH(veProdMsgDefConst);
	device.veProdMsgDefVolatile = msgMemory;
	device.veProdMsgDefInvolatile = msgSettings;
	device.signalDefinition = veProdSgnDef;
	device.fpCounters = fpCounters;

	device.nadItem = nad;
	veItemCtx(nad)->ptr = &device;
	veItemSetChanged(nad, nadSettingChanged);

	veCanAddSendingDevice(&device);

	/* If the nad is valid, go on-bus directly, otherwise wait for it to become valid */
	nadSettingChanged(nad);
}

/**
 * When other devices are used for sending, this should be called when done,
 * so the mainDevice is normally always used.
 */
void mainDeviceActivate(void)
{
	j1939DevActivate(&device);
}
