#include <velib/velib_inc_J1939.h>

#include <velib/vecan/prod_data.h>
#include <velib/vecan/prod_data_cm.h>
#include <velib/J1939/J1939_device.h>
#include <velib/utils/ve_assert.h>

#include <senders.h>

J1939DeviceTp *sendingDevices;
static J1939DeviceTp *lastDevice;

/*
 * Most of the time there is only one device sending data towards the CAN-bus and
 * often the messages it send are predefined and that where the producer interface
 * is based upon. On a color control etc, the devices / messages depend on the
 * devices known by the color control and might be found later etc. To make such
 * dynamic devices possible, without much code changes, a per device offset is simply
 * substracted from the msgNr, so there is a per device msgNr starting at 0.
 */

void veCanAddSendingDevice(struct J1939Device *device)
{
	device->next = NULL;
	if (sendingDevices == NULL) {
		device->prod.msgFirst = 0;
		device->prod.msgLast = device->msgCount - 1;
		sendingDevices = device;
		lastDevice = device;
	} else {
		device->prod.msgFirst = lastDevice->prod.msgLast + 1;
		device->prod.msgLast = device->prod.msgFirst + device->msgCount - 1;
		lastDevice->next = device;
		lastDevice = device;
	}
}

VeMsgNr veProdMessageNumber(void)
{
	if (lastDevice == NULL)
		return 0;
	return lastDevice->prod.msgLast + 1;
}

void veProdScaledValueCallback(VeProdSgnDef const *signal, VeVariant *variant)
{
	if (j1939Device.veProdScaledCb)
		j1939Device.veProdScaledCb(signal, variant);
}

un8* veProdSequenceCallback(VeMsgNr nr)
{
	veAssert(nr >= j1939Device.prod.msgFirst && nr <= j1939Device.prod.msgLast);
	return &j1939Device.fpCounters[nr - j1939Device.prod.msgFirst];

}

veBool veProdSendMsgEvent(VeMsgNr nr)
{
	VE_UNUSED(nr);
	return veTrue;
}


VeProdMsgDefConst const *getProdMsgConst(VeMsgNr id)
{
	veAssert(id >= j1939Device.prod.msgFirst && id <= j1939Device.prod.msgLast);
	return &j1939Device.veProdMsgDefConst[id - j1939Device.prod.msgFirst];
}


static J1939DeviceTp *cachedDevice;

/**
 * @brief lookup the device which sends the message with given id
 * @param id
 * @return the device sending this message
 *
 * note: since mostly if not always, the messages are iterated the last
 * used device is checked first, since most likely that is the correct one.
 * note: if devices are ever freed, don t forget to reset this as well!
 */
J1939DeviceTp *findDevice(VeMsgNr id)
{
	if (cachedDevice && id >= cachedDevice->prod.msgFirst && id <= cachedDevice->prod.msgLast)
		return cachedDevice;

	cachedDevice = sendingDevices;
	while (cachedDevice) {
		if (id <= cachedDevice->prod.msgLast)
			break;
		cachedDevice = cachedDevice->next;
	}

	veAssert(cachedDevice && id >= cachedDevice->prod.msgFirst && id <= cachedDevice->prod.msgLast);
	return cachedDevice;
}

VeProdMsgDefVolatile *getProdMsgVolatile(VeMsgNr id)
{
	J1939DeviceTp *dev = findDevice(id);
	veAssert(dev && id >= dev->prod.msgFirst && id <= dev->prod.msgLast);
	return &dev->veProdMsgDefVolatile[id - dev->prod.msgFirst];
}

VeProdMsgDefInvolatile *getProdMsgInvolatile(VeMsgNr id)
{
	J1939DeviceTp *dev = findDevice(id);
	veAssert(dev && id >= dev->prod.msgFirst && id <= dev->prod.msgLast);
	return &dev->veProdMsgDefInvolatile[id - dev->prod.msgFirst];
}


VeProdSgnDef const *veProdGetSignalDef(VeMsgNr msgNr)
{
	int msgIndex = msgNr - j1939Device.prod.msgFirst;
	int sgnNr = j1939Device.veProdMsgDefConst[msgIndex].signal;
	return &j1939Device.signalDefinition[sgnNr];
}

veBool veCanIsLocalSender(struct J1939Func *listeningDev)
{
	J1939DeviceTp *dev = sendingDevices;

	/* Update the device claimed here */
	while (dev) {
		if (dev->nad == listeningDev->nmt.nad)
			return veTrue;
		dev = dev->next;
	}

	return veFalse;
}


/*
 *  note: only changing instances it supported at the moment. If more settings
 * are supported this should be changed, but is fine for now.
 */
VeProdSgnSpanDef const def = {0, 0xFF};

/// Gets the minimum and maximum allowed values for signals which can be set.
VeProdSgnSpanDef const* veProdSgnSpanCallback(VeProdSgnDef const * signal)
{
	VE_UNUSED(signal);

	return &def;
}

/**
 * @brief
 *	Callback to indicate the start of the change.
 */
void veProdDataChangeBeginEvent(VeMsgNr msgNr)
{
	VE_UNUSED(msgNr);
}

/**
 * @brief
 *	Callback to indicate that a request to update signal to the value in variant
 *	has arrived.
 * @details
 *	The values is aleady checked for sanity and min / max. The
 *	application must perform the actual update or return an error code. The error
 *	code will be send back to the source of the command.
  */
N2kCmParamCode veProdDataChangeEvent(VeProdSgnDef const* signal, VeVariant* variant)
{
	// Update memory
	veProdDataCmSet(signal, variant);

	return N2K_CM_PARAM_NO_ERROR;
}

/**
 * Callback to indicate that at least one signal is changed and the processing of
 * the command is done.
 */
void veProdDataStoreEvent(void)
{
	if (j1939Device.storeSettings)
		j1939Device.storeSettings(&j1939Device);
}
