#include <stdlib.h>
#include <string.h>

#include <velib/velib_inc_J1939.h>

#include <velib/J1939/J1939_device.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_fp.h>
#include <velib/vecan/prod_types.h>

#include <senders.h>

#define sptr(a)		(void*)(&(a)),sizeof(a)
#define ptr(a)		(void*)(&(a))
#define fptr(a)		(void*)(&(a)),0
#define strptr(a)	(void*)(&(a[0])),sizeof(a)
#define nodata		NULL, 0

typedef enum
{
	MAP_UNITARY,
	MAP_DC_V_FROM_ITEM
} veScalesNames;

// NMEA200
static un16	const n2kDb				= 1301;								/* NMEA 2000 db version */
static un16	const n2kProductID		= 0xFFFF;							/* Product id */
static char	const n2kModel[]		= "vlink";							/* Model */
static char	const n2kSwVersion[]	= "1.0";							/* Software version */
static char	const n2kModelVersion[] = "1.0";							/* Model Version */
static un8	const n2kCert			= 1;								/* Certification level */
static un8	const n2kLoad			= 0;								/* Load equivalant (50mA) */
static un16 const victron			= 0x9966;							/* proprietary code */
static un16 const hub1Vreg			= VE_REG_LINK_VSET;
static un32 const zeropad = 0;

static un16 converterStateVreg = VE_REG_DEVICE_STATE;
static un8 const converterState = N2K_CONVERTER_STATE_HUB1;

static un8 const pgnTxOpcode = 0;
static un8 const pgnRxOpcode = 1;

#define PGN(a)	(a&0xFF),(un8)(a>>8)

static un8 const pgnsRx[] =
{
	// data page 0, addressable
	PGN(J1939_PGN0_REQUEST),		0,
	PGN(J1939_PGN0_ACL),			0,

	// data page 1, addressable
	PGN(N2K_PGN1_COMPLEX_GROUP),	1,
};

static un8 const pgnsTx[] =
{
	// data page 0, addressable
	PGN(J1939_PGN0_ACK), 			0,
	PGN(J1939_PGN0_ACL), 			0,

	// data page 1, addressable
	PGN(N2K_PGN1_COMPLEX_GROUP), 	1,
	PGN(N2K_PGN1_PGN_LIST), 		1,

	// data page 1, broadcast
	PGN(N2K_PGN1_PRODUCT_INFO), 	1,
};


static VeProdSgnDef * getDef(VLinkDevice *hub1)
{
	VeProdSgnDef *ret;
	J1939DeviceTp *dev = &hub1->j1939Dev;

	VeProdSgnDef const def[] =
	{
#define N2K_PRODUCT_INFO_S			0
#define N2K_PRODUCT_INFO_C			8
#define N2K_PRODUCT_INFO_E			(N2K_PRODUCT_INFO_S + N2K_PRODUCT_INFO_C)

		{VE_UN16,		MAP_UNITARY,			0,						sptr(n2kDb)									},
		{VE_UN16,		MAP_UNITARY,			0,						sptr(n2kProductID)							},
		{VE_CHAR32,		MAP_UNITARY,			0,						strptr(n2kModel)							},
		{VE_CHAR32,		MAP_UNITARY,			0,						strptr(n2kSwVersion)						},
		{VE_CHAR32,		MAP_UNITARY,			0,						strptr(n2kModelVersion)						},
		{VE_CHAR32,		MAP_UNITARY,			0,						strptr("serial#")							},
		{VE_UN8,		MAP_UNITARY,			0,						sptr(n2kCert)								},
		{VE_UN8,		MAP_UNITARY,			0,						sptr(n2kLoad)								},

#define ACL_S	N2K_PRODUCT_INFO_E
#define ACL_C	10
#define ACL_E	(ACL_S + ACL_C)
		{VE_BIT21,		MAP_UNITARY,			0,						ptr(dev->name.low)		,	0				},
		{VE_BIT11,		MAP_UNITARY,			0,						ptr(dev->name.low)		,	21				},
		{VE_BIT3,		MAP_UNITARY,			S_WRITE,				ptr(dev->name.high)		,	0				},
		{VE_BIT5,		MAP_UNITARY,			S_WRITE,				ptr(dev->name.high)		,	3				},
		{VE_BIT8,		MAP_UNITARY,			0,						ptr(dev->name.high)		,	8				},
		{VE_BIT1,		MAP_UNITARY,			0,						ptr(dev->name.high)		,	16				},
		{VE_BIT7,		MAP_UNITARY,			0,						ptr(dev->name.high)		,	17				},
		{VE_BIT4,		MAP_UNITARY,			0,						ptr(dev->name.high)		,	24				},
		{VE_BIT3,		MAP_UNITARY,			0,						ptr(dev->name.high)		,	28				},
		{VE_BIT1,		MAP_UNITARY,			0,						ptr(dev->name.high)		,	31				},

#define VE_REG_LINK_VSET_S ACL_E
#define VE_REG_LINK_VSET_C 4
#define VE_REG_LINK_VSET_E (VE_REG_LINK_VSET_S + VE_REG_LINK_VSET_C)
		{VE_UN16,		MAP_UNITARY,			0,						sptr(victron)								},
		{VE_UN16,		MAP_UNITARY,			0,						sptr(hub1Vreg)								},
		{VE_UN16,		MAP_DC_V_FROM_ITEM,		0,						hub1->linkVoltageSetPoint,		0					},
		{VE_UN16,		MAP_UNITARY,			0,						sptr(zeropad)								},


#define VE_REG_DEVICE_STATE_S VE_REG_LINK_VSET_E
#define VE_REG_DEVICE_STATE_C 4
#define VE_REG_DEVICE_STATE_E (VE_REG_DEVICE_STATE_S + VE_REG_DEVICE_STATE_C)
		{VE_UN16,		MAP_UNITARY,			0,						sptr(victron)								},
		{VE_UN16,		MAP_UNITARY,			0,						sptr(converterStateVreg)					},
		{VE_UN8,		MAP_UNITARY,			0,						sptr(converterState)						},
		{VE_UN24,		MAP_UNITARY,			0,						sptr(zeropad)								},

#define N2K_PGNS_TX_S				VE_REG_DEVICE_STATE_E
#define N2K_PGNS_TX_C				2
#define N2K_PGNS_TX_E				(N2K_PGNS_TX_S + N2K_PGNS_TX_C)
		{VE_UN8,		MAP_UNITARY,			S_INST,					sptr(pgnTxOpcode)							},
		{VE_BUF,		MAP_UNITARY,			0,						sptr(pgnsTx)								},

#define N2K_PGNS_RX_S				N2K_PGNS_TX_E
#define N2K_PGNS_RX_C				2
#define F0_N2K_PGNS_RX_E				(N2K_PGNS_RX_S + N2K_PGNS_RX_C)
		{VE_UN8,		MAP_UNITARY,			S_INST,					sptr(pgnRxOpcode)							},
		{VE_BUF,		MAP_UNITARY,			0,						sptr(pgnsRx)								},

	};

	ret = malloc(sizeof(def));
	if (!ret)
		return NULL;
	memcpy(ret, def, sizeof(def));
	return ret;
}

static VeProdMsgDefConst const veProdMsgDefConst[] =
{
	{N2K_FP,	1,	N2K_PGN1_PRODUCT_INFO,		N2K_PRODUCT_INFO_S,		N2K_PRODUCT_INFO_C},
	{J1939_SF,	0,	J1939_PGN0_ACL,				ACL_S,					ACL_C},
	{J1939_SF,  0,	N2K_PGN1_PROPRIETARY_PDU1,	VE_REG_LINK_VSET_S,		VE_REG_LINK_VSET_C},
	{J1939_SF,  0,	N2K_PGN1_PROPRIETARY_PDU1,	VE_REG_DEVICE_STATE_S,	VE_REG_DEVICE_STATE_C},
	{N2K_FP,	1,	N2K_PGN1_PGN_LIST,			N2K_PGNS_TX_S,			N2K_PGNS_TX_C},
	{N2K_FP,	1,	N2K_PGN1_PGN_LIST,			N2K_PGNS_RX_S,			N2K_PGNS_RX_C},
};

static  VeProdMsgDefInvolatile msgSettings[ARRAY_LENGTH(veProdMsgDefConst)] = {
	{0,		6},
	{0,		6},
	{20,	7},
	{20,	7},
	{0,		6},
	{0,		6},
};

static void veProdScaledValueCallback(VeProdSgnDef const *signal, VeVariant *variant)
{
	VeItem *item;
	VeVariant v;

	switch (signal->trans)
	{
	case MAP_DC_V_FROM_ITEM:
		item = signal->value;
		if (veItemIsValid(item))
			veVariantUn16(variant, veItemValue(item, &v)->value.Float * 100);
		else
			veVariantUn16(variant, 0xFFFF);
		break;
	default:
		;
	}
}

/** Announce the device on the dbus after all settings are valid. */
static void announceDeviceWhenComplete(VLinkDevice *vLink)
{
	if (vLink->j1939Dev.isAnnounced)
		return;

	if (veItemIsValid(vLink->j1939Dev.instanceItem) && veItemIsValid(vLink->j1939Dev.nadItem)) {
		vLink->j1939Dev.isAnnounced = veTrue;
		veCanAddSendingDevice(&vLink->j1939Dev);
		j1939AclEnable();
	}
}

static void nadSettingChanged(VeItem *item)
{
	VeVariant v;
	VLinkDevice *vLink = (VLinkDevice *) veItemCtx(item)->ptr;
	j1939DevActivate(&vLink->j1939Dev);
	un8 nad = veItemValue(item, &v)->value.UN8;

	if (!vLink->j1939Dev.isAnnounced)
	{
		vLink->j1939Dev.nad = nad;
		announceDeviceWhenComplete(vLink);
	} else {
		if (nad != j1939Device.nad)
			j1939AclChangeNad(nad);
	}

	mainDeviceActivate();
}

static void instanceSettingChanged(VeItem *item)
{
	VeVariant v;
	VLinkDevice *vLink = (VLinkDevice *) veItemCtx(item)->ptr;
	j1939DevActivate(&vLink->j1939Dev);

	vLink->j1939Dev.name.high.vun32 &= ~0xFF;
	vLink->j1939Dev.name.high.vun32 |= veItemValue(item, &v)->value.UN8;

	if (!vLink->j1939Dev.isAnnounced)
		announceDeviceWhenComplete(vLink);
	else
		j1939AclUpdateState(J1939_ACL_EV_REQ);

	mainDeviceActivate();
}

static void storeSettings(J1939DeviceTp *dev)
{
	VeVariant v;

	veItemSet(dev->instanceItem, veVariantUn8(&v, N2K_ACL_H_GET_DEVICE_INST(dev->name.high)));
}

/**
 * In order to allow hub1 to work without a vebus CAN-bus cable, the behaviour of
 * the cable is mimicked. Pretend to be a multi and send the setpoint from that
 * device. Since it is needed for hub4 as well, call it vlink instead of hub1.
 *
 * @param unique		the unique number for the ACL
 * @param instance		the instance number to use
 * @param nad			the network address to be used as default
 * @param hub1Setpoint	the setpoint to forward to the CAN-bus
 *
 * Note: the change events of above items are implemented, so when these the values
 * afterwards, their changes make it to the CAN-bus. Likewise if their value changes
 * on the CAN-bus they are updated accordingly.
 */
VLinkDevice *vLinkNewInstance(un32 unique, VeItem *instance, VeItem *nad, VeItem *hub1Setpoint)
{
	VLinkDevice *vLink = calloc(1, sizeof(VLinkDevice));
	J1939DeviceTp *device = &vLink->j1939Dev;

	if (device == NULL)
		return NULL;

	vLink->j1939Dev.nadItem = nad;
	veItemCtx(nad)->ptr = vLink;
	veItemSetChanged(nad, nadSettingChanged);

	vLink->j1939Dev.instanceItem = instance;
	veItemCtx(instance)->ptr = vLink;
	veItemSetChanged(instance, instanceSettingChanged);

	vLink->linkVoltageSetPoint = hub1Setpoint;

	device->veProdScaledCb = veProdScaledValueCallback;
	device->storeSettings = storeSettings;
	device->name.low =  J1939_ACL_L_MAN(N2K_MAN_VICTRON) | J1939_ACL_L_IDENTITY_NUMBER(unique);
	device->name.high.vun32 = J1939_ACL_H_INDUSTRY(J1939_INDUSTRY_MARINE) | J1939_ACL_H_SCA |
			J1939_ACL_H_CLASS(N2K_CLASS_ELECTRICAL_GENERATION) | J1939_ACL_H_FUNCTION(N2K_FUNCTION_INVERTER);
	j1939DevActivate(device);

	j1939AclInit();

	device->veProdMsgDefConst = veProdMsgDefConst;
	device->msgCount = ARRAY_LENGTH(veProdMsgDefConst);
	device->veProdMsgDefVolatile = calloc(device->msgCount, sizeof(VeProdMsgDefInvolatile));
	device->veProdMsgDefInvolatile = msgSettings;
	device->signalDefinition = getDef(vLink);
	if (device->signalDefinition == NULL)
		return NULL;
	device->fpCounters = calloc(device->msgCount, 1);

	nadSettingChanged(nad);
	instanceSettingChanged(nad);

	mainDeviceActivate();

	return vLink;
}
