#include <velib/velib_inc_J1939.h>

#include <velib/io/ve_stream.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/J1939/J1939.h>
#include <velib/J1939/J1939_function.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_complex.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/products.h>
#include <velib/vecan/reg.h>
#include <velib/vecan/reg_rx.h>

#include <msg_router.h>
#include <values_battery.h>
#include <vecan_common.h>

#define INVALID_INSTANCE 0xFF
#define ALARM 2

static void veInDcStatus(DcStatus *status)
{
	VeVariant v;

	veInN2kDcV(&status->voltage);
	veInN2kDcI(&status->current);

	/* Calculate DC power if both current and voltage are valid */
	if (veItemIsValid(&status->voltage) && veItemIsValid(&status->current)) {
		VeVariant *varV = &status->voltage.variant;
		VeVariant *varI = &status->current.variant;
		veAssert(varV->type.tp == VE_FLOAT && varI->type.tp == VE_FLOAT);
		veItemOwnerSet(&status->power, veVariantFloat(&v, varV->value.Float * varI->value.Float));
	} else {
		veItemInvalidate(&status->power);
	}
}

static void veInRelayAndAlarms(un16 productId, BatteryAlarms *alarms)
{
	if (productId != VE_PROD_ID_VECAN_LYNX_SHUNT) {
		// Bmv alarms
		veInUn8(); // instance
		veInItemLe(&alarms->relayState, VE_BIT2, 0);
		veInItemLe(&alarms->alarm, VE_BIT2, 0);
		veInAlarm(&alarms->lowVoltage);
		veInAlarm(&alarms->highVoltage);
		veInAlarm(&alarms->lowSoc);
		veInAlarm(&alarms->lowStarterVoltage);
		veInAlarm(&alarms->highStarterVoltage);
	} else {
		// Lynx shunt alarms
		VeVariant v;
		un8 instance = veInUn8(); // instance

		if (instance == 0) {
			veInAlarm(&alarms->lowVoltage);
			veInAlarm(&alarms->highVoltage);
			veInAlarm(&alarms->lowSoc);
			veInAlarm(&alarms->lowFusedVoltage);
			veInAlarm(&alarms->highFusedVoltage);
			veInAlarm(&alarms->fuseBlown);
			veInAlarm(&alarms->highBatteryTemperature);
			veInAlarm(&alarms->lowBatteryTemperature);
			veInAlarm(&alarms->highInternalTemperature);
			if ((alarms->lowVoltage.variant.value.UN32 == ALARM) ||
				(alarms->highVoltage.variant.value.UN32 == ALARM) ||
				(alarms->lowSoc.variant.value.UN32 == ALARM) ||
				(alarms->lowFusedVoltage.variant.value.UN32 == ALARM) ||
				(alarms->highFusedVoltage.variant.value.UN32 == ALARM) ||
				(alarms->fuseBlown.variant.value.UN32 == ALARM) ||
				(alarms->highBatteryTemperature.variant.value.UN32 == ALARM) ||
				(alarms->lowBatteryTemperature.variant.value.UN32 == ALARM) ||
				(alarms->highInternalTemperature.variant.value.UN32 == ALARM))
				veItemOwnerSet(&alarms->alarm, veVariantUn8(&v, 1));
			else
				veItemOwnerSet(&alarms->alarm, veVariantUn8(&v, 0));
		}
	}
}

static void veInTime(VeItem* item)
{
	un16 value;
	un32 timeInSeconds;
	VeVariant v;

	value = veInUn16Le();
	if (n2kStreamIn.error || value == 0xFFFF || value == 0xFFFE) {
		veItemInvalidate(item);
	}
	else {
		timeInSeconds = value * 60;
		veItemOwnerSet(item, veVariantUn32(&v, timeInSeconds));
	}
}

static void veInLowSocAlarmWrongScale(VeItem *item)
{
	VeVariant variant;
	un16 value;

	value = veInUn16Le();
	if (n2kStreamIn.error) {
		veItemInvalidate(item);
		return;
	}

	/*
	 * Only allow value 0-9, which corresponds with 0-90%.
	 * If a faulty value is detected, it is corrected and send to the device.
	 * The device will respond with the 'correct' value which will be processed in this function again.
	 */
	if (value > 9) {
		veItemSet(item, veVariantFloat(&variant, value * 0.1));
		return;
	}

	/* Convert to % and set*/
	veItemOwnerSet(item, veVariantFloat(&variant, 10 * value));
}

static void handleMessage(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeBattery *battery = (VeBattery*) ctx;

	// by default message are only processed once and then freed.
	canId = n2kMsgIn->msgId;

	// Remove target address for device specific messages.
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
	{
		// handle incoming message for the zero data page.
		switch(J1939PgnFromId(canId))
		{
		default:
			break;
		}
	} else {
		// handle incoming message for the first data page
		switch (J1939PgnFromId(canId))
		{
		case N2K_PGN1_DC_DETAILS:
			{
				un8 instance, dcType;
				veInUn8(); // SID
				instance = veInUn8();
				dcType = veInUn8();
				if (dcType == DC_TYPE_BATTERY)
				{
					VeVariant v;
					un8 soc = veInUn8();

					battery->batteryInstance = instance;

					if (validUN8(soc))
						veItemOwnerSet(&battery->stateOfCharge, veVariantFloat(&v, soc));
					else
						veItemInvalidate(&battery->stateOfCharge);

					veInUn8(); // SOH
					veInTime(&battery->timeToGo);
				}
			}
			break;

		case N2K_PGN1_BATTERY_STATUS:
			{
				if (battery->batteryInstance != INVALID_INSTANCE)
				{
					un8 instance = veInUn8();
					if (instance != battery->batteryInstance) {
						veInN2kDcV(&battery->dcStatusAux.voltage);
					}
					else { // BMV only sends one DC_DETAILS and not more than two BATTERY_STATUS.
						veInDcStatus(&battery->dcStatusBattery);
						veInTemperatureInCelsius(&battery->batteryTemperature);
					}
				}
			}
			break;

		case N2K_PGN1_SWITCH_STATUS:
		case N2K_PGN1_SWITCH_CONTROL:
			if (func->system && func->system->master == func) {
				veInRelayAndAlarms(func->system->product.productId.variant.value.UN16, &battery->alarms);
			}
			break;

		default:
			break;
		}
	}
}

/*=== proprietary messages from the vebus system ===*/
static void handleVreg(void *ctx, struct J1939Func *src, VeEvent ev,
							VeRegId id,	VeRegAckCode code)
{
	VeBattery *battery = (VeBattery *) ctx;
	VeVariant var;
	VE_UNUSED(src);
	VE_UNUSED(code);

	switch (ev)
	{
	case VE_EV_VREG_RX_ACK:
		switch ((int)id) {
		case VE_REG_RESTORE_DEFAULT:
			if (code != VACK_OK)
				break;
			veItemValue(&battery->hasSettings, &var);
			/* Request all settings */
			if (veItemIsValid(&battery->hasSettings))
				battery->can.getSettings = var.value.UN32 != 0;
			break;
		default:
			;
		}
		break;

	case VE_EV_VREG_RX:
		switch ((int) id)
		{
		case VE_REG_HIST_DEEPEST_DISCHARGE:
		case VE_REG_HIST_LAST_DISCHARGE:
		case VE_REG_HIST_AVERAGE_DISCHARGE:
		case VE_REG_HIST_CUMULATIVE_AH:
			veInItemLe(&battery->history.h[id - VE_REG_HIST_DEEPEST_DISCHARGE], VE_SN32, 1);
			break;

		case VE_REG_HIST_NR_OF_CHARGE_CYCLES:
		case VE_REG_HIST_NR_OF_FULL_DISCHARGES:
		case VE_REG_HIST_SECS_SINCE_LAST_FULL_CHARGE:
		case VE_REG_HIST_NR_OF_AUTO_SYNCS:
		case VE_REG_HIST_NR_OF_LOW_VOLTAGE_ALARMS:
		case VE_REG_HIST_NR_OF_HIGH_VOLTAGE_ALARMS:
		case VE_REG_HIST_NR_OF_LOW_VOLTAGE_2_ALARMS:
		case VE_REG_HIST_NR_OF_HIGH_VOLTAGE_2_ALARMS:
			veInItemLe(&battery->history.h[id - VE_REG_HIST_DEEPEST_DISCHARGE], VE_SN32, 0);
			break;

		case VE_REG_HIST_MIN_VOLTAGE:
		case VE_REG_HIST_MAX_VOLTAGE:
		case VE_REG_HIST_MIN_VOLTAGE_2:
		case VE_REG_HIST_MAX_VOLTAGE_2:
			veInItemLe(&battery->history.h[id - VE_REG_HIST_DEEPEST_DISCHARGE], VE_SN32, 2);
			break;

		case VE_REG_HIST_KWH_OUT:
		case VE_REG_HIST_KWH_IN:
			veInItemLe(&battery->history.h[id - VE_REG_HIST_DEEPEST_DISCHARGE], VE_UN32, 2);
			break;

		case VE_REG_BMV_CE:
			veInItemLe(&battery->consumedAmphours, VE_SN32, 1);
			break;

		case VE_REG_ALARM_LOW_SOC_SET:
		case VE_REG_ALARM_LOW_SOC_CLEAR:
			if (battery->hasLowSocAlarmBug) {
				if (id == VE_REG_ALARM_LOW_SOC_SET) {
					veInLowSocAlarmWrongScale(&battery->settings.alarm.lowSoc);
				} else if (id == VE_REG_ALARM_LOW_SOC_CLEAR) {
					veInLowSocAlarmWrongScale(&battery->settings.alarm.lowSocClear);
				}
				break;
			}
			/* Fall-thru */

		default:
			/* Only read settings when supported */
			veItemValue(&battery->hasSettings, &var);
			if (var.value.UN32 == 0)
				break;
			veInSetting(battery->settingsCtx, id, battery);

			switch (id) {
			case VE_REG_LOW_SOC:
				/*
				 * Applies to Lynx Shunt < v1.09:
				 * Manually set the relay lowSoc as it is linked to the discharge floor, but separate items on the dbus
				 */
				if (battery->dischargeFloorLinkedToRelay.variant.value.UN32)
					veItemOwnerSet(&battery->settings.relay.lowSoc, &battery->settings.battery.dischargeFloor.variant);
				break;
			case VE_REG_LOW_SOC_CLEAR:
				/* Applies to Lynx Shunt < v1.09: see comment @ VE_REG_LOW_SOC */
				if (battery->dischargeFloorLinkedToRelay.variant.value.UN32)
					veItemOwnerSet(&battery->settings.relay.lowSocClear, &battery->settings.battery.lowSocClear.variant);
				break;
			default:
				;
			}
			break;
		}
		break;
	default:
		break;
	}
}

static void checkSettingsSupport(VeSystem *system)
{
	VeBattery *battery = (VeBattery *)system->canbusCtx;
	VeDevVup* vup = NULL;
	VeDevVup* vupItt;
	VeVariant hasSettings;
	VeVariant dischargeFloorLinkedToRelay;
	un16 version;

	/* Get the last vup instance. E.g. 0 for Lynx Shunt and 1 for BMV @ CAN interface */
	un8 i = 0;
	while ((vupItt = veDevUdfFindVup(system->master, i)) != NULL) {
		vup = vupItt;
		i++;
	}

	if (!vup) return;

	/* Only use lower 2 bytes. 3rd byte indicates test version or release candidate */
	version = vup->appVersion & 0xFFFF;

	veVariantUn32(&hasSettings, version >= battery->minSettingsVersion);
	veVariantUn32(&dischargeFloorLinkedToRelay, version < battery->minRelayLowSocVersion);

	/* Get settings if has changed to 'supported' */
	if (!veVariantIsEqual(&battery->hasSettings.variant, &hasSettings))
		battery->can.getSettings = hasSettings.value.UN32 != 0;

	veItemOwnerSet(&battery->hasSettings, &hasSettings);
	veItemOwnerSet(&battery->dischargeFloorLinkedToRelay, &dischargeFloorLinkedToRelay);

	/* Low Soc alarm bug: off by factor 100 */
	battery->hasLowSocAlarmBug = vup->prodId == VE_PROD_ID_VECAN_LYNX_SHUNT && version <= 0x108;
}

static void update(VeSystem *system)
{
	VeBattery *battery = (VeBattery *) system->canbusCtx;

	if (battery->can.getHistory)
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_HIST_DEEPEST_DISCHARGE, 0xFFE0))
			battery->can.getHistory = veFalse;

	if (battery->can.getRelayInfo)
		if (veRegRequestWhenOnBus(system->master, VE_REG_RELAY_CONTROL))
			battery->can.getRelayInfo = veFalse;
	/*
	 * Request product-info because there are situation like the BMV-to-N2K-cable without BMV.
	 * The update is done in vecan_common.
	 */
	if (battery->can.getProductInfo) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_PRODUCT_ID) &&
				veRegRequestWhenOnBus(system->master, VE_REG_APP_VER)) {
			veDevN2kReq(system->master, VE_DEV_N2K_PROD_INFO);
			battery->can.getProductInfo = veFalse;
		}
	}

	/* Check settings support */
	if (battery->can.checkSettings) {
		battery->can.checkSettings = veFalse;
		checkSettingsSupport(system);
	}

	/* Battery settings are from 0x1000 - 0x1034, alarm settings are from 0x320 - 0x332
	 * and relay settings are from 0x34D - 0x362 */
	if (battery->can.getSettings)
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_BATTERY_CAPACITY, 0xFFC0) &&
			veRegRequestMaskWhenOnBus(system->master, VE_REG_ALARM_LOW_VOLTAGE_SET, 0xFF80))
			battery->can.getSettings = veFalse;

	/* Call the childs tick function as it is not in the system list */
	if (system->child && system->child->update)
		system->child->update(system->child);

	/* In case of Lynx Shunt, the battery settings need to be locked, as they
	 * are controlled by the Lynx Ion. NB: Not for Lynx Ion + Shunt. */
	/* Note: no need to use veDevUdfFindVup as both product only have 1 vup instance */
	if (battery->can.pollBms && system->master->udf.vup->prodId == VE_PROD_ID_VECAN_LYNX_SHUNT) {
		J1939FuncTp *itt = j1939FuncList;
		while (itt) {
			if (itt->udf.vup && itt->udf.vup->prodId == VE_PROD_ID_VECAN_LYNX_ION) {
				VeVariant v;
				veItemOwnerSet(&battery->settings.battery.locked, veVariantUn32(&v, 1));
				break;
			}
			itt = itt->next;
		}
		battery->can.pollBms = veFalse;
	}
}

static void tick(VeSystem *system)
{
	VeBattery *battery = (VeBattery *) system->canbusCtx;

	if (battery->can.pollTimer == 0 || --battery->can.pollTimer == 0)
	{
		battery->can.getHistory = veTrue;
		battery->can.getProductInfo = veTrue;
		battery->can.getRelayInfo = veTrue;
		battery->can.pollBms = veTrue;
		battery->can.checkSettings = veTrue;
		battery->can.pollTimer = 6 /* sec */ * 20;
	}

	/* Call the childs tick function as it is not in the system list */
	if (system->child && system->child->tick)
		system->child->tick(system->child);
}

veBool veCanBatteryAdd(VeSystem *system)
{
	VeBattery *battery = (VeBattery *) system->canbusCtx;

	/* If it's a sub system, use the can hooks as
	 * the master's hook is allready in use by the parent system */
	MsgRouterHook *hooks = system->parent ?
				&battery->can.hooks : &system->master->hooks;

	hooks->rxFun = handleMessage;
	hooks->vregFun = handleVreg;
	hooks->ctx = battery;
	hooks->func = system->master;
	system->update = update;
	system->tick = tick;

	battery->can.pollTimer = 0;
	battery->can.getHistory = veTrue;
	battery->can.getProductInfo = veTrue;
	battery->can.getRelayInfo = veTrue;
	battery->can.checkSettings = veTrue;

	battery->batteryInstance = INVALID_INSTANCE;

	msgRouterAddHook(hooks);

	return veTrue;
}
