#include <velib/velib_inc_J1939.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/vecan/charger_error.h>
#include <velib/vecan/products.h>
#include <velib/io/ve_stream_items.h>
#include <velib/utils/ve_logger.h>

#include <msg_router.h>
#include <values_charger.h>
#include <vecan_charger.h>
#include <vecan_common.h>
#include <vecan_charger.h>

#include <stdlib.h>

#define MODULE	"charger"

static MsgRouterHook hook;

veBool veCanChargerSetMode(VeSystem *system, sn32 mode)
{
	if (!veRegPrepareSf(system->master->nmt.nad, VE_REG_DEVICE_MODE))
		return veFalse;

	veOutUn8(mode);
	veOutUn24Le(0);
	veOutSendSf();

	return veTrue;
}

veBool veCanChargerSetCurrentLimit(VeSystem *system, un16 limit)
{
	if (!veRegPrepareSf(system->master->nmt.nad, VE_REG_AC_IN_1_CURRENT_LIMIT_REMOTE))
		return veFalse;

	veOutUn16Le(limit);
	veOutUn16Le(0);
	veOutSendSf();

	return veTrue;
}

static void veInDcStatusAndTemperature(DcStatus *status, VeItem *temperature)
{
	veInN2kDcV(&status->voltage);
	veInN2kDcI(&status->current);
	veInTemperatureInCelsius(temperature);
}

static void veInPvStatus(DcStatus *status)
{
	veInN2kDcV(&status->voltage);
	veInN2kDcI(&status->current);
}

static void veInRelayAndAlarms(Alarms *alarms)
{
	veInUn8(); /* instance */
	veInItemLe(&alarms->relayState, VE_BIT2, 0);
	veInAlarm(&alarms->alarm);
	veInAlarm(&alarms->lowVoltage);
	veInAlarm(&alarms->highVoltage);
}

/* handle grid charger NMEA2000 PGNs */
static void handleGridMsg(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeCharger *charger = (VeCharger *)ctx;

	VE_UNUSED(func);

	/* Do not update fields if we have a config error */
	if (charger->can.flags & VE_CAN_LINK_CONFIG_ERROR) {
		return;
	}

	/* by default message are only processed once and then freed. */
	canId = n2kMsgIn->msgId;

	/* Remove target address for device specific messages. */
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	/* Skip the zero data page */
	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
		return;

	/* handle incoming message for the first data page */
	switch (J1939PgnFromId(canId))
	{
	case N2K_PGN1_BATTERY_STATUS:
		{
			un8 instance = veInUn8();

			if (instance < ARRAY_LENGTH(charger->dcStatus)) {
				veInN2kDcV(&charger->dcStatus[instance].voltage);
				veInN2kDcI(&charger->dcStatus[instance].current);

				/* Instance 0 contains the battery temperature */
				if (instance == 0) {
					veInTemperatureInCelsius(&charger->temperature);
				}
			}
		}
		break;

	case N2K_PGN1_AC_POWER_CURRENT_PHASE_A:
		veInUn8(); /* skip sid */
		veInUn8(); /* skip connection number */
		veInItemLe(&charger->acIn.current, VE_UN16, 1);
		veInItemLe(&charger->acIn.power, VE_SN32, 0);
		break;

	default:
		;
	}
}

/* handle solar charger NMEA2000 PGNs */
static void handleSolarMsg(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeCharger *charger = (VeCharger *)ctx;

	VE_UNUSED(func);

	/* Do not update fields if we have a config error */
	if (charger->can.flags & VE_CAN_LINK_CONFIG_ERROR) {
		return;
	}

	/* by default message are only processed once and then freed. */
	canId = n2kMsgIn->msgId;

	/* Remove target address for device specific messages. */
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	/* Skip the zero data page */
	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
		return;

	/* handle incoming message for the first data page */
	switch (J1939PgnFromId(canId))
	{
	case N2K_PGN1_DC_DETAILS:
		{
			un8 instance, dcType;

			veInUn8(); /* skip Sequence ID */
			instance = veInUn8();
			dcType = veInUn8();
			switch (dcType)
			{
			case DC_TYPE_BATTERY:
				charger->can.batteryInstance = instance;
				break;

			case DC_TYPE_SOLAR_CELL:
				charger->can.solarCellInstance = instance;
				break;

			default:
				;
			}
		}
		break;

	case N2K_PGN1_BATTERY_STATUS:
		{
			un8 instance = veInUn8();

			if (instance == charger->can.batteryInstance) {
				veInDcStatusAndTemperature(&charger->dcStatus[0], &charger->temperature);
			} else if (instance == charger->can.solarCellInstance) {
				/* Pv voltage and current only makes sense in stand alone mode */
				if (charger->can.standalone) {
					veInPvStatus(&charger->pvStatus);
				}
			}
		}
		break;

	default:
		;
	}
}

static void handleMessage(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeCharger *charger = (VeCharger *)ctx;
	VeVariant variant, error;

	VE_UNUSED(func);

	/* Do not update fields if we have a config error */
	if (charger->can.flags & VE_CAN_LINK_CONFIG_ERROR) {
		return;
	}

	/* by default message are only processed once and then freed. */
	canId = n2kMsgIn->msgId;

	/* Remove target address for device specific messages. */
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
	{
		/* handle incoming message for the zero data page. */
		switch(J1939PgnFromId(canId))
		{
		default:
			break;
		}
	} else {
		/* handle incoming message for the first data page */
		switch (J1939PgnFromId(canId))
		{
		/* Deprecated, but older version still use it */
#if 0
		case N2K_PGN1_CHARGER_STATUS:
			veInUn8(); /* skip charger instance */
			veInUn8(); /* skip battery instance */
			veInUn8(); /* skip charger state and mode */
			veInItemLe(&charger->mode, VE_BIT2, 0);
			veInItemLe(&charger->equalization.pending, VE_BIT2, 0);
			veInBits8Le(4); /* skip reserved */
			veInItemLe(&charger->equalization.timeRemaining, VE_UN16, 0);
			break;
#endif

		case N2K_PGN1_CONVERTER_STATUS:
			veInUn8(); /* sequence id */
			veInUn8(); /* connection number */
			veInVariantDecimalsLe(&variant, VE_UN8, 0);

			/*
			 * Check for suppressed errors, then set state from fault to off.
			 * MPPTs < v2.03 need to suppress error 19 (reversed polarity).
			 */
			veItemValue(&charger->errorCode, &error);
			if (error.value.UN8 == CHARGER_ERROR_NONE) {
				if (variant.value.UN8 == N2K_CONVERTER_STATE_FAULT)
					veVariantUn8(&variant, N2K_CONVERTER_STATE_OFF);
			}

			veItemOwnerSet(&charger->state, &variant);
			break;

		case N2K_PGN1_SWITCH_STATUS:
			/* Suppress during switching of master. */
			/* Old firmware triggers an error 66 when the master election is not yet completed. */
			if (charger->can.suppressTimer)
				break;
			veInRelayAndAlarms(&charger->alarms);
			break;

		default:
			;
		}
	}
}

static void handleVreg(void *ctx, struct J1939Func *src, VeEvent ev,
									VeRegId id, VeRegAckCode code)
{
	VeCharger *charger = (VeCharger*) ctx;
	VeVariant variant;

	VE_UNUSED(src);
	VE_UNUSED(code);

	/* Do not update fields if we have a config error */
	if (charger->can.flags & VE_CAN_LINK_CONFIG_ERROR) {
		return;
	}

	switch (ev)
	{
	case VE_EV_VREG_RX:
		switch ((int) id)
		{
		case VE_REG_DC_INPUT_POWER:
			veInItemLe(&charger->yield.power, VE_UN32, 2);
			break;
		case VE_REG_CHR_SYSTEM_YIELD:
			veInItemLe(&charger->yield.system, VE_UN32, 2);
			break;
		case VE_REG_CHR_USER_YIELD:
			veInItemLe(&charger->yield.user, VE_UN32, 2);
			break;
		case VE_REG_CHR_TODAY_YIELD:
			veInItemLe(&charger->hist[0].yield, VE_UN32, 2);
			break;
		case VE_REG_CHR_TODAY_PMAX:
			veInItemLe(&charger->hist[0].maxPower, VE_UN16, 0);
			break;
		case VE_REG_CHR_YESTERDAY_YIELD:
			veInItemLe(&charger->hist[1].yield, VE_UN32, 2);
			break;
		case VE_REG_CHR_YESTERDAY_PMAX:
			veInItemLe(&charger->hist[1].maxPower, VE_UN16, 0);
			break;
		case VE_REG_CHR_ERROR_CODE:
			veInVariantDecimalsLe(&variant, VE_UN8, 0);

			switch (variant.value.UN8)
			{
			/* Workaround for Skyllas < v2.04 and MPPTs < v2.03.
			 * These versions trow an error 66 (incompatible device) during
			 * the master election when the config query is still pending */
			case CHARGER_ERROR_LINK_CONFIGURATION:
				if (!veVariantIsEqual(&charger->can.tmpError, &variant))
					charger->can.suppressTimer = 5 /* seconds */ * 20;
				break;

			/* Always suppress: MPPTs < v2.03.
			 * This has no effect on the Skyllas as they don't have this error. */
			case CHARGER_ERROR_CHARGER_CURRENT_REVERSED:
				veVariantUn8(&variant, CHARGER_ERROR_NONE);
				break;

			default:
				charger->can.suppressTimer = 0;
			}

			charger->can.tmpError = variant;

			/* As long as the error is suppressed, assign its own value to prevent invalidation */
			if (charger->can.suppressTimer)
				veItemValue(&charger->errorCode, &variant);

			veItemOwnerSet(&charger->errorCode, &variant);
			break;
		case VE_REG_DEVICE_MODE:
			veInItemLe(&charger->mode, VE_UN8, 0);
			break;
		case VE_REG_REMOTE_CONTROL_USED:
			if ((veInUn32Le() & charger->can.remoteControl) != charger->can.remoteControl)
				charger->can.flags |= VE_CAN_SEND_RC_USED;
			break;
		case VE_REG_AC_IN_1_CURRENT_LIMIT:
			veInItemLe(&charger->acInCurrentLimit, VE_UN16, 1);
			break;
		case VE_REG_AC_IN_1_CURRENT_LIMIT_MIN:
			veInVariantDecimalsLe(&variant, VE_UN16, 1);
			veItemSetMin(&charger->acInCurrentLimit, &variant);
			break;
		case VE_REG_AC_IN_1_CURRENT_LIMIT_MAX:
			veInVariantDecimalsLe(&variant, VE_UN16, 1);
			veItemSetMax(&charger->acInCurrentLimit, &variant);
			break;
		case VE_REG_CHR_OUTPUTS:
			veInItemLe(&charger->outputs, VE_UN8, 0);
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
}

static void update(VeSystem *system)
{
	VeCharger *charger = (VeCharger *) system->canbusCtx;

	/* request yields and maximum charge power (today and yesterday). */
	if (charger->can.flags & VE_CAN_REQ_YESTERDAY_PMAX_GROUP)
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_CHR_YESTERDAY_PMAX, 0xFFFC))
			charger->can.flags &= ~VE_CAN_REQ_YESTERDAY_PMAX_GROUP;

	/* request yields system and user. */
	if (charger->can.flags & VE_CAN_REQ_VE_REG_CHR_USER_YIELD)
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_CHR_USER_YIELD, 0xFFFE))
			charger->can.flags &= ~VE_CAN_REQ_VE_REG_CHR_USER_YIELD;

	/* request the error code. */
	if (charger->can.flags & VE_CAN_REQ_VE_REG_ERROR_CODE)
		if (veRegRequestWhenOnBus(system->master, VE_REG_CHR_ERROR_CODE))
			charger->can.flags &= ~VE_CAN_REQ_VE_REG_ERROR_CODE;

	/* check if the charger is standalone. */
	if (charger->can.flags & VE_CAN_CHECK_STANDALONE)
	{
		if (veCanCountDevices(system) == 1)
			charger->can.standalone = veTrue;
		else
			charger->can.standalone = veFalse;
		charger->can.flags &= ~VE_CAN_CHECK_STANDALONE;
	}

	/* Remote control: on/off, current limit, etc */
	if (charger->can.flags & VE_CAN_SEND_RC_USED) {
		if (veRegPrepareSfWhenOnBus(system->master, VE_REG_REMOTE_CONTROL_USED)) {
			veOutUn32Le(charger->can.remoteControl);
			veOutSendSf();

			charger->can.flags &= ~VE_CAN_SEND_RC_USED;
		}
	}

	/* AC in current limit */
	if (charger->can.flags & VE_CAN_REQ_VE_REG_AC_CURRENT_LIMIT) {
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_AC_IN_1_CURRENT_LIMIT, 0xFFF8))
			charger->can.flags &= ~VE_CAN_REQ_VE_REG_AC_CURRENT_LIMIT;
	}

	/* PV power */
	if (charger->can.flags & VE_CAN_REQ_VE_REG_DC_INPUT_POWER) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_DC_INPUT_POWER))
			charger->can.flags &= ~VE_CAN_REQ_VE_REG_DC_INPUT_POWER;
	}

	/* Nr of outputs */
	if (charger->can.flags & VE_CAN_REQ_VE_REG_CHR_OUTPUTS) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_CHR_OUTPUTS))
			charger->can.flags &= ~VE_CAN_REQ_VE_REG_CHR_OUTPUTS;
	}
}

static void tick(VeSystem *system)
{
	VeCharger *charger = (VeCharger *)system->canbusCtx;

	if (charger->can.pollTimer == 0  || --charger->can.pollTimer == 0) {
		charger->can.flags |= VE_CAN_CHECK_STANDALONE | VE_CAN_REQ_VE_REG_ERROR_CODE;

		/* In case of solar charger: request yield, etc */
		if (system->master->udf.group == VE_GROUP_SOLAR_CHARGERS)
			charger->can.flags |= VE_CAN_REQ_YESTERDAY_PMAX_GROUP | VE_CAN_REQ_VE_REG_CHR_USER_YIELD;

		charger->can.pollTimer = 5 /* sec */ * 20;
	}

	if (charger->can.suppressTimer)
		charger->can.suppressTimer--;

	/* Count the number of masters. If it is more than 1 for more than 5 seconds,
	 * force a configuration error (error 66). The user needs to configure the system. */
	if (veCanCountMasters(system) > 1) {
		if (charger->can.configErrorTimer && --charger->can.configErrorTimer == 0) {

			/* Invalidate items without time out */
			veItemInvalidate(&charger->yield.power);
			veItemInvalidate(&charger->yield.user);
			veItemInvalidate(&charger->yield.system);

			charger->can.flags |= VE_CAN_LINK_CONFIG_ERROR;
			logE(system->id, "found multiple masters -> force configuration error");
		}
	} else {
		charger->can.configErrorTimer = 5 /* seconds */ * 20;
		charger->can.flags &= ~VE_CAN_LINK_CONFIG_ERROR;
	}

	/* Force error link configuration error */
	if (charger->can.flags & VE_CAN_LINK_CONFIG_ERROR) {
		VeVariant variant;

		veVariantUn8(&variant, CHARGER_ERROR_LINK_CONFIGURATION);
		veItemOwnerSet(&charger->errorCode, &variant);
	}
}

veBool veCanChargerCheckMaster(VeSystem *system, struct J1939Func *dev)
{
	VeCharger *charger = (VeCharger *)system->canbusCtx;

	/* Needs to be in the same group */
	if (system->master->udf.group != dev->udf.group)
		return veFalse;

	if (dev->system) {
		/* Check if current master has become master of another system */
		veBool changedToOtherGroup = dev->system == system &&
			system->instance != N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high);

		/* Remove empty systems. If there are still devices in the
		 * system, one of them will become the new master. */
		if (veCanCountDevices(dev->system) == 0) {
			logI(system->id, "switching master: %u -> ?", dev->nmt.nad);
			veCanRemoveCharger(dev->system);
		}

		if (changedToOtherGroup) {
			msgRouterRemoveHook(&dev->hooks);

			/* Remove charger context */
			dev->hooks.ctx = NULL;
			return veFalse;
		}
	}

	/* The device instance need to be equal */
	if (system->instance !=	N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high))
		return veFalse;

	logI(system->id, "switching master: %u -> %u", system->master->nmt.nad, dev->nmt.nad);

	/* Unregister hook to prevent that we're listening to the old master */
	if (system->instance == N2K_ACL_H_GET_DEVICE_INST(system->master->nmt.name.high))
		msgRouterRemoveHook(&system->master->hooks);

	/* Also unregister the new master as this could be the old master of another
	 * system. E.g. when the instance has changed from x to y back to x */
	msgRouterRemoveHook(&dev->hooks);

	/* Copy the msg hooks from the current master.
	 * The next pointer is allready set to NULL */
	dev->hooks.rxFun = system->master->hooks.rxFun;
	dev->hooks.vregFun = system->master->hooks.vregFun;
	dev->hooks.func = dev;
	dev->hooks.ctx = system->canbusCtx;
	dev->system = system;

	/* Set new master */
	system->master = dev;
	charger->can.hooks.func = system->master;

	msgRouterAddHook(&system->master->hooks);

	/* Request master info */
	veDevUdfReq(system->master);
	veDevN2kReq(system->master, VE_DEV_N2K_PROD_INFO);

	/* Request some regs as they are not broadcasted periodically */
	if (system->master->udf.group == VE_GROUP_GRID_CHARGERS) {
		charger->can.flags |= VE_CAN_REQ_VE_REG_AC_CURRENT_LIMIT;
	}

	if (system->master->udf.group == VE_GROUP_SOLAR_CHARGERS) {
		charger->can.flags |= VE_CAN_REQ_YESTERDAY_PMAX_GROUP |
				VE_CAN_REQ_VE_REG_CHR_USER_YIELD | VE_CAN_REQ_VE_REG_DC_INPUT_POWER;
	}

	/* Nr of outputs can change depending on the master */
	charger->can.flags |= VE_CAN_REQ_VE_REG_CHR_OUTPUTS;

	return veTrue;
}

veBool veCanChargerAdd(VeSystem *system)
{
	VeCharger *charger = (VeCharger *)system->canbusCtx;

	msgRouterRemoveHook(&system->master->hooks);
	msgRouterRemoveHook(&charger->can.hooks);

	system->master->hooks.rxFun = handleMessage;
	system->master->hooks.vregFun = handleVreg;
	system->master->hooks.ctx = charger;
	system->master->hooks.func = system->master;
	system->update = update;
	system->tick = tick;
	system->instance = N2K_ACL_H_GET_DEVICE_INST(system->master->nmt.name.high);
	charger->can.batteryInstance = 0xFF;
	charger->can.solarCellInstance = 0xFF;
	charger->can.pollTimer = 1 /* sec */ * 20;
	charger->can.standalone = veFalse;
	charger->can.suppressTimer = 0;
	charger->can.configErrorTimer = 5 /* seconds */ * 20;

	/* note: this can be tweaked on the actual product, but is only done once */
	charger->can.flags = VE_CAN_SEND_RC_USED | VE_CAN_REQ_VE_REG_ERROR_CODE |
			VE_CAN_CHECK_STANDALONE | VE_CAN_REQ_VE_REG_CHR_OUTPUTS;

	/* Hook for common messages */
	msgRouterAddHook(&system->master->hooks);

	switch (system->master->udf.group)
	{
	case VE_GROUP_GRID_CHARGERS:
		/* Only grid chargers have rc support for now */
		charger->can.remoteControl = VE_REG_RC_ON_OFF_MODE | VE_REG_RC_AC1_IN_LIMIT;
		charger->can.flags |= VE_CAN_REQ_VE_REG_AC_CURRENT_LIMIT;
		charger->can.hooks.ctx = charger;
		charger->can.hooks.rxFun = handleGridMsg;
		charger->can.hooks.func = system->master;
		msgRouterAddHook(&charger->can.hooks);
		break;

	case VE_GROUP_SOLAR_CHARGERS:
		charger->can.flags |= VE_CAN_REQ_YESTERDAY_PMAX_GROUP |
				VE_CAN_REQ_VE_REG_DC_INPUT_POWER | VE_CAN_REQ_VE_REG_CHR_USER_YIELD;
		charger->can.hooks.ctx = charger;
		charger->can.hooks.rxFun = handleSolarMsg;
		charger->can.hooks.func = system->master;
		msgRouterAddHook(&charger->can.hooks);
		break;

	default:
		;
	}

	return veTrue;
}

VeSystem *veCanRemoveCharger(VeSystem *system)
{
	VeCharger *charger = (VeCharger *)system->canbusCtx;

	msgRouterRemoveHook(&charger->can.hooks);
	msgRouterRemoveHook(&system->master->hooks);

	free(charger);

	return veCanRemoveSystem(system);
}

/* look for new chargers / master taking over */
static void handleMsg(void *ctx, struct J1939Func *func)
{
	VE_UNUSED(ctx);

	if (J1939MsgNrFromId(n2kMsgIn->msgId) != N2K_MSGNR_CONVERTER_STATUS)
		return;

	if (J1939_ACL_L_GET_MAN(func->nmt.name.low) == N2K_MAN_VICTRON &&
			veCanIsElectricalFunction(func, N2K_FUNCTION_CHARGER))
	{
		/* Increments if converter status is not received anymore */
		func->isMaster = 3;

		veCanCheckForNewDevice(func);
	}
}

void veCanChargersInit(void)
{
	hook.rxFun = handleMsg;
	msgRouterAddHook(&hook);
}
