#include <stdlib.h>
#include <string.h>

#include <velib/velib_inc_J1939.h>

#include <velib/io/ve_stream_n2k.h>
#include <velib/J1939/J1939.h>
#include <velib/J1939/J1939_function.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_complex.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/reg_rx.h>
#include <velib/vecan/dev_udf.h>
#include <velib/vecan/products.h>

#include <msg_router.h>
#include <vecan_common.h>

veBool veCanIsElectricalFunction(struct J1939Func *func, N2kFunctionElectrical function)
{
	J1939NameHighFields *fields = &func->nmt.name.high.fields;

	return fields->indGroup == J1939_INDUSTRY_MARINE &&
		fields->devClass == N2K_CLASS_ELECTRICAL_GENERATION &&
		fields->function == function;
}

void veInVregFirmwareVersion(VeItem *item)
{
	veInItemLe(item, VE_UN24, 0);
}

void veInProductId(VeItem *item)
{
	veInItemLe(item, VE_UN16, 0);
}

static void veConvertTemperatureKelvinToCelsius(VeItem *item)
{
	VeVariant *v = &item->variant;
	if (veVariantIsValid(v))
	{
		veVariantToFloat(v);
		v->value.Float = v->value.Float - 273.15F;
	}
}

void veInTemperatureInCelsius(VeItem *item)
{
	static VeItem tmp;

	/*
	 * To prevent a false dbus-property-changed use a local VeItem,
	 * because veInN2kDcT calls veItemOwnerSet.
	 */
	veInN2kDcT(&tmp);
	veConvertTemperatureKelvinToCelsius(&tmp);
	veItemOwnerSet(item, &tmp.variant);
}

/*
 * Read the N2K alarm and convert values to dbus-alarm.
 * N2K alarm; 0 = no alarm, 1 = alarm.
 * dbus alarm; 0 = no alarm, 1 = warning, 2 = alarm.
 */
void veInAlarm(VeItem *item)
{
	static VeItem tmp;

	/*
	 * To prevent a false dbus-property-changed use a local VeItem.
	 */
	veInItemLe(&tmp, VE_BIT2, 0);
	if (tmp.variant.value.UN32 == 1)
		tmp.variant.value.UN32 = 2;
	veItemOwnerSet(item, &tmp.variant);
}

void veInSetting(VeSetting *settings, VeRegId id, void *src)
{
	VeSetting *setting = settings;

	while (setting && setting->id != VE_REG_INVALID) {
		if (setting->id == id) {
			VeItem *item = (VeItem *)((un8 *)src + setting->offset);

			VeVariant variant;
			variant.type.tp = setting->type;
			veInVariantLe(&variant);

			if (setting->scale)
				veVariantTransformIn(&variant, setting->scale);

			veItemOwnerSet(item, &variant);
			return;
		}

		setting++;
	}
}

veBool veCanSetDeviceInstance(struct J1939Func *dev, un8 instance)
{
	VeVariant v;

	/* Prepare a message. */
	n2kCmAllocCmd(dev->nmt.nad, 0, J1939_PGN0_ACL);
	if (n2kMsgOut == NULL)
		return veFalse;
	n2kCmParam(3, veVariantUn8(&v, instance & 0x07));
	n2kCmParam(4, veVariantUn8(&v, (instance & 0xF8) >> 3));
	n2kCmSend();

	return veTrue;
}

veBool veCanSetVeReg(VeSystem *system, VeVariant *variant, VeSetting *setting)
{
	if (!system->master)
		return veFalse;

	if (setting->isLocked && setting->isLocked(system->canbusCtx))
		return veFalse;

	if (!veRegPrepareSf(system->master->nmt.nad, setting->id))
		return veFalse;

	veVariantToFloat(variant);

	if (setting->scale)
		veVariantTransformOut(variant, setting->scale);

	veVariantFromFloat(variant, setting->type);
	veOutVariantLe(variant);
	veOutSendSf();

	return veTrue;
}

void j1939FuncNew(J1939FuncTp *func)
{
	/* get basic info, product id / version etc */
	veDevUdfReq(func);
	veDevN2kReq(func, VE_DEV_N2K_PROD_INFO);

	veCanNewDeviceCallback(func);
}

void msgRouterInstanceChanged(J1939FuncTp *func)
{
	/*
	 * This could mean that a new system is created or systems have merged, i.e.
	 * the master has changed. The j1929FuncNew is not triggered, because the device
	 * was allready known.
	 */
	veCanCheckForNewDevice(func);
	veCanInstanceChangedCallback(func);
}

J1939FuncTp* j1939FuncAlloc(J1939Name* name)
{
	/* Victron products always */
	if (J1939_ACL_L_GET_MAN(name->low) != N2K_MAN_VICTRON) {

		/* Offshore systems tanks sensor */
		if (name->high.fields.indGroup != J1939_INDUSTRY_MARINE || name->high.fields.devClass != N2K_CLASS_GENERAL)
			return NULL;
	}

	return (J1939FuncTp*) calloc(1, sizeof(J1939FuncTp));
}

void veMsgHandlerVeRegs(VeEvent ev)
{
	J1939FuncTp *func;

	if (ev != VE_EV_VREG_RX)
		return;

	func = j1939FuncByNad(n2kMsgIn->msgId.p.src);
	if (!func)
		return;

	switch (veRegRxId) {
	case VE_REG_GROUP_ID:
		veCanCheckForNewDevice(func);
		break;

	default:
		;
	}
}

/*
 * BMV to NMEA2K cable return a version like 0x20C.
 * Which should be 0x212.
 * This function corrects that misstake.
 */
static un32 correctVersion(un32 version)
{
	un8 byte = version & 0xF;
	un32 newVersion = version;

	if (byte > 0x9)
	{
		newVersion += 6;
	}

	return newVersion;
}

/*
 * Get the serial number from number.
 * Number must be a null-terminated string.
 * Return the string after the first space.
 * If no space found, number is returned.
 */
static char * getSerialNumber(char *number)
{
	char *p = number;
	un32 i;
	veBool found = veFalse;

	for (i = 0; (i < strlen(number)) && (found == veFalse); i++)
	{
		if (*p == ' ')
			found = veTrue;
		p++;
	}

	if (found == veTrue)
		return p;
	else
		return NULL;
}

static void itemSet(VeSystem *system, VeItem *item, VeVariant *variant)
{
	if (!system || !item)
		return;

	veItemOwnerSet(item, variant);
	itemSet(system->child, (VeItem *)item->ctx.ptr, variant);
}

static void deviceTick1Sec(J1939FuncTp *dev)
{
	VeDevVup* vup = NULL;
	VeDevVup* vupItt;
	VeVariant v;
	un8 instance = 0;
	VeSystem *sys = dev->system;

	/* Used to indicate if dev is a master. Only product info of the master is updated/shown */
	veBool isMaster = sys && sys->master == dev;

	// Get the last vup instance.
	while ((vupItt = veDevUdfFindVup(dev, instance)) != NULL) {
		vup = vupItt;
		instance++;
	}

	if (isMaster) {
		itemSet(sys, &sys->product.connected, veVariantUn32(&v,
			!((dev->nmt.flags & J1939_NMT_DEV_DISCONNECTED) || (vup && vup->prodId == VE_PROD_NOT_SET))));
		itemSet(sys, &sys->product.instance, veVariantUn8(&v, N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high)));
	}

	if (vup)
	{
		if (vup->flags & VE_DEV_VUP_PROD_ID)
		{
			vup->flags &= ~VE_DEV_VUP_PROD_ID;

			if (dev->needsProductId) {
				veCanCheckForNewDevice(dev);
				dev->needsProductId = veFalse;
			}

			if (isMaster) {
				if (veVariantIsValid(veVariantUn16(&v, vup->prodId))) {
					itemSet(sys, &sys->product.productId, &v);

					if (sys->product.ptrProductName) {
						itemSet(sys, sys->product.ptrProductName, veVariantStr(&v, veProductGetName(vup->prodId)));
					}
				}
			}
		}

		if (vup->flags & VE_DEV_VUP_VER)
		{
			vup->flags &= ~VE_DEV_VUP_VER;

			if (isMaster) {
				if (vup->appVersion != VREG_VER_INVALID_VERSION) {
					if ((vup->prodId >= VE_PROD_ID_BMV600S) && (vup->prodId <= VE_PROD_ID_BMV700H))
					{
						VeDevVup *vupCable = veDevUdfFindVup(dev, 0);
						if (vupCable) {
							if (vupCable->appVersion <= 0x106)
								itemSet(sys, &sys->product.firmwareVersion, veVariantUn32(&v, correctVersion(vup->appVersion)));
							else
								itemSet(sys, &sys->product.firmwareVersion, veVariantUn32(&v, vup->appVersion));
						}
					} else {
						itemSet(sys, &sys->product.firmwareVersion, veVariantUn32(&v, vup->appVersion));
					}
				}
			}
		}
	}

	if (dev->n2k.valid & VE_DEV_N2K_PROD_INFO) {
		veItemOwnerSet(&dev->n2kProductName, veVariantStr(&v, dev->n2k.prod.modelId));

		if (isMaster) {
			itemSet(sys, &sys->product.serial, veVariantStr(&v, getSerialNumber(dev->n2k.prod.serial)));

			/* Non-vup */
			if (vup == NULL) {
				itemSet(sys, &sys->product.firmwareVersion, veVariantStr(&v, dev->n2k.prod.firmwareVersion));
				itemSet(sys, &sys->product.hardwareVersion, veVariantStr(&v, dev->n2k.prod.modelVersion));
			}
		}
	}

	/* isMaster is used as counter to check if there are
	 * multiple masters in a system (which is not ok) */
	if (dev->isMaster)
		dev->isMaster--;
}

void veCanbusCommonTick(void)
{
	static un8 sec;
	J1939FuncTp *dev;

	if (sec != 0 && --sec != 0)
		return;

	sec = 20;

	dev = j1939FuncList;
	while (dev) {
		deviceTick1Sec(dev);
		dev = dev->next;
	}
}

un32 decimalToHex(un32 version)
{
	un32 hex = 0;
	un32 factor = 1;

	while (version != 0) {
		hex += (version % 16) * factor;
		version /= 16;
		factor *= 10;
	}

	return hex;
}
