#include <velib/velib_inc_J1939.h>

#include <velib/io/ve_stream.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/J1939/J1939.h>
#include <velib/J1939/J1939_function.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_complex.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/products.h>
#include <velib/vecan/reg.h>
#include <velib/vecan/reg_rx.h>

#include <msg_router.h>
#include <values_lynxion.h>
#include <vecan_common.h>

static VeVariant v;

static un8 flagToNumber(un32 flags)
{
	un8 i;
	un8 j = 0;
	if (flags) {
		for (i = 0; i < 31; i++) {
			if (flags & 1) {
				j = i;
				break;
			}
			flags = flags >> 1;
		}
	}
	return j;
}

static void handleMessage(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeLynxIon* lynxIon = (VeLynxIon*) ctx;
	VE_UNUSED(func);

	// by default message are only processed once and then freed.
	canId = n2kMsgIn->msgId;

	// Remove target address for device specific messages.
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
	{
		// handle incoming message for the zero data page.
		switch(J1939PgnFromId(canId))
		{
		default:
			break;
		}
	} else {
		// handle incoming message for the first data page
		switch (J1939PgnFromId(canId))
		{
		case N2K_PGN1_BATTERY_STATUS:
		{
			un8 instance = veInUn8();
			if (instance == 1)
				veInItemLe(&lynxIon->minCellVoltage, VE_SN16, 2);
			else if (instance == 2)
				veInItemLe(&lynxIon->maxCellVoltage, VE_SN16, 2);
			break;
		}
		default:
			break;
		}
	}
}

/*=== prorietary messages from the vebus system ===*/
static void handleVreg(void *ctx, struct J1939Func *src, VeEvent ev,
							VeRegId id,	VeRegAckCode code)
{
	VeLynxIon *lynxIon = (VeLynxIon *) ctx;
	VE_UNUSED(src);
	VE_UNUSED(code);

	switch (ev)
	{
	case VE_EV_VREG_RX:
		switch ((int) id)
		{
		case VE_REG_LYNX_ION_BMS_STATE:
			veInItemLe(&lynxIon->state, VE_UN8, 0);
			break;
		case VE_REG_LYNX_ION_BMS_ERROR_FLAGS:
		{
			un32 flags = veInUn32Le();
			veItemOwnerSet(&lynxIon->error, veVariantUn8(&v, flagToNumber(flags)));
			break;
		}
		case VE_REG_LYNX_ION_FLAGS:
		{
			un32 ionFlags = veInUn32Le();
			veItemOwnerSet(&lynxIon->systemSwitch, veVariantUn8(&v, (ionFlags & VE_VDATA_ION_FLG_RELAY_DISCHARGED) ? 1 : 0));
			veItemOwnerSet(&lynxIon->balancing, veVariantUn8(&v, (ionFlags & VE_VDATA_ION_FLG_BALANCE) ? 1 : 0));
			veItemOwnerSet(&lynxIon->allowToCharge, veVariantUn8(&v, (ionFlags & VE_VDATA_ION_FLG_CHARGED) ? 0 : 1));
			veItemOwnerSet(&lynxIon->allowToDischarge, veVariantUn8(&v, (ionFlags & VE_VDATA_ION_FLG_DISCHARGED) ? 0 : 1));
			veItemOwnerSet(&lynxIon->externalRelay, veVariantUn8(&v, (ionFlags & VE_VDATA_ION_FLG_RELAY_DISCHARGED) ? 1 : 0));
			break;
		}
		case VE_REG_BATTERY_CONFIGURATION:
			veInItemLe(&lynxIon->nrOfBatteries, VE_UN8, 0);
			veInItemLe(&lynxIon->nrOfCellsPerBattery, VE_UN8, 0);
			veInItemLe(&lynxIon->batteriesParallel, VE_UN8, 0);
			veInItemLe(&lynxIon->batteriesSeries, VE_UN8, 0);
			break;
		case VE_REG_LYNX_ION_SHUTDOWNS_DUE_ERROR:
			veInItemLe(&lynxIon->shutdownsDueError, VE_UN16, 0);
			break;
		case VE_REG_LYNX_ION_LAST_ERROR:
			veInItemLe(&lynxIon->lastError1, VE_UN8, 0);
			veInItemLe(&lynxIon->lastError2, VE_UN8, 0);
			veInItemLe(&lynxIon->lastError3, VE_UN8, 0);
			veInItemLe(&lynxIon->lastError4, VE_UN8, 0);
			break;
		case VE_REG_BATTERY_CELL_VOLTAGE_MIN_MAX:
			veInItemLe(&lynxIon->history.minCellVoltage, VE_UN16, 2);
			veInItemLe(&lynxIon->history.maxCellVoltage, VE_UN16, 2);
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
}

static void update(VeSystem *system)
{
	VeLynxIon* lynxIon = (VeLynxIon *) system->canbusCtx;

	/* Min/max cell voltage, non-volatile */
	if (lynxIon->can.getHistory) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_BATTERY_CELL_VOLTAGE_MIN_MAX))
			lynxIon->can.getHistory = veFalse;
	}

	/* BMS state, flags and error */
	if (lynxIon->can.getBmsInfo) {
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_LYNX_ION_FLAGS, 0xFFFC))
			lynxIon->can.getBmsInfo = veFalse;
	}

	/* Diagnostics: last error, nr of shut downs due error */
	if (lynxIon->can.getDiagnostics) {
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_LYNX_ION_SHUTDOWNS_DUE_ERROR, 0xFFFC))
			lynxIon->can.getDiagnostics = veFalse;
	}

	/* Battery system */
	if (lynxIon->can.getSystem) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_BATTERY_CONFIGURATION))
			lynxIon->can.getSystem = veFalse;
	}

	/* Request product-info */
	if (lynxIon->can.getProductInfo) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_PRODUCT_ID) &&
				veRegRequestWhenOnBus(system->master, VE_REG_APP_VER)) {
			veDevN2kReq(system->master, VE_DEV_N2K_PROD_INFO);
			lynxIon->can.getProductInfo = veFalse;
		}
	}

	/* Call the childs update function as it is not in the system list */
	if (system->child && system->child->update)
		system->child->update(system->child);
}

static void tick(VeSystem *system)
{
	VeLynxIon* lynxIon = (VeLynxIon *) system->canbusCtx;

	if (lynxIon->can.pollTimer == 0 || --lynxIon->can.pollTimer == 0)
	{
		lynxIon->can.getBmsInfo = veTrue;
		lynxIon->can.getHistory = veTrue;
		lynxIon->can.getDiagnostics = veTrue;
		lynxIon->can.getSystem = veTrue;
		lynxIon->can.pollTimer = 6 /* sec */ * 20;
	}

	/* Call the childs tick function as it is not in the system list */
	if (system->child && system->child->tick)
		system->child->tick(system->child);
}

veBool veCanLynxIonAdd(VeSystem *system)
{
	VeLynxIon *lynxIon = (VeLynxIon *) system->canbusCtx;

	/* If it's a sub system, use the can hooks as
	 * the master's hook is allready in use by the parent system */
	MsgRouterHook *hooks = system->parent ?
				&lynxIon->can.hooks : &system->master->hooks;

	hooks->rxFun = handleMessage;
	hooks->vregFun = handleVreg;
	hooks->ctx = lynxIon;
	hooks->func = system->master;
	system->tick = tick;
	system->update = update;

	lynxIon->can.pollTimer = 0;
	lynxIon->can.getProductInfo = veTrue;
	lynxIon->can.getBmsInfo = veTrue;
	lynxIon->can.getHistory = veTrue;
	lynxIon->can.getDiagnostics = veTrue;
	lynxIon->can.getSystem = veTrue;

	msgRouterAddHook(hooks);

	return veTrue;
}
