#include <velib/velib_inc_J1939.h>

#include <velib/io/ve_stream.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/J1939/J1939.h>
#include <velib/J1939/J1939_function.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_complex.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/products.h>

#include "msg_router.h"
#include "values_tank.h"
#include "vecan_common.h"

#include <stdio.h>

#define INVALID_INSTANCE 0xFF

static MsgRouterHook hook;

/*
 * Setting the tank capacity works/tested for the following devices:
 * - Victron VE.Can resistive tank sender adaptor (Offshore Systems UK Ltd 3125 v3.02V)
 * - Offshore Systems 3125 Tank Sender Adaptor < v3.0.2, with workaround for endianess issue
 * - Offshore Systems 3125 Tank Sender Adaptor >= v3.0.2
 * - Maretron TLA100 - Tank Level Adaptor
 */
veBool veCanTankSetCapacity(VeSystem *system, un32 capacity)
{
	VeTank *tank = (VeTank *)system->canbusCtx;
	un8 *paramCount;

	switch (system->master->n2k.prod.code) {
	/* Workaround for Offshore Systems 3125 Tank Sender Adaptor < v3.0.2. Complex command has wrong endianess. */
	case 961: {
		un32 v[3];
		if (sscanf(system->master->n2k.prod.firmwareVersion ,"v%x.%x.%x", &v[2], &v[1], &v[0]) != 3)
			return veFalse;

		un32 version = (v[2] << 16) | (v[1] << 8) | v[0];
		if (version < 0x030002) {
			capacity = ((capacity & 0x0000FFFF) << 16) | ((capacity & 0xFFFF0000) >> 16);
			capacity = ((capacity & 0x00FF00FF) <<  8) | ((capacity & 0xFF00FF00) >>  8);
		}
		break;
	}
	default:
		;
	}

	paramCount = n2kCmAllocCmd(system->master->nmt.nad, 1, N2K_PGN1_FLUID_LEVEL);

	if (n2kMsgOut == NULL)
		return veFalse;

	/* Parameter field 1: tank instance */
	veOutUn8(1);
	veOutUn8(tank->tankInstance);

	/* Parameter field 4: capacity */
	veOutUn8(4);
	veOutUn32Le(capacity);

	*paramCount = 2;
	n2kCmSend();

	return veTrue;
}

/*
 * Setting the fluid works for the following devices:
 * - Maretron TLA100 - Tank Level Adaptor
 *
 * The following devices have a physical switch to select the fluid type,
 * therefor setting the setting the fluid type through NMEA2000 has no effect
 * - Victron VE.Can resistive tank sender adaptor
 * - Offshore Systems 3125 Tank Sender Adaptor
 */
veBool veCanTankSetFluidType(VeSystem *system, un8 type)
{
	VeTank *tank = (VeTank *)system->canbusCtx;
	un8 *paramCount;

	paramCount = n2kCmAllocCmd(system->master->nmt.nad, 1, N2K_PGN1_FLUID_LEVEL);

	if (n2kMsgOut == NULL)
		return veFalse;

	/* Parameter field 1: tank instance */
	veOutUn8(1);
	veOutUn8(tank->tankInstance);

	/* Parameter field 2: fluid type */
	veOutUn8(2);
	veOutUn8(type);

	*paramCount = 2;
	n2kCmSend();

	return veTrue;
}

static void veInN2kFluidType(VeItem *item)
{
	VeVariant variant;

	veVariantUn8(&variant, veInBits8Le(4));

	if (!veVariantIsValid(&variant) || n2kStreamIn.error) {
		veVariantInvalidate(&variant);
	}

	veItemOwnerSet(item, &variant);
}

static void veInN2kFluidLevel(VeItem *item)
{
	VeVariant variant;

	veVariantSn16(&variant, veInSn16Le());

	if (!veVariantIsValid(&variant) || n2kStreamIn.error) {
		veVariantInvalidate(&variant);
	} else {
		veVariantToFloat(&variant);
		variant.value.Float *= 0.004;
	}

	veItemOwnerSet(item, &variant);
}

static void veInN2kFluidCapacity(VeItem *item)
{
	VeVariant variant;

	veVariantUn32(&variant, veInUn32Le());

	if (n2kStreamIn.error) {
		veVariantInvalidate(&variant);
	}
	/* Check for Reserved/Error/Not availabe */
	else if (variant.value.UN32 < 0xFFFFFFFD) {
		veVariantToFloat(&variant);
		variant.value.Float /= 10000;
	}
	else {
		veVariantInvalidate(&variant);
	}

	veItemOwnerSet(item, &variant);
}

static void handleMessage(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeTank* tank = (VeTank*) ctx;

	VE_UNUSED(func);

	// by default message are only processed once and then freed.
	canId = n2kMsgIn->msgId;

	// Remove target address for device specific messages.
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
	{
		// handle incoming message for the zero data page.
		switch(J1939PgnFromId(canId))
		{
		default:
			break;
		}
	} else {
		// handle incoming message for the first data page
		switch (J1939PgnFromId(canId))
		{
		case N2K_PGN1_FLUID_LEVEL:
			tank->tankInstance = veInBits8Le(4);
			veInN2kFluidType(&tank->fluidType);
			veInN2kFluidLevel(&tank->level);
			veInN2kFluidCapacity(&tank->capacity);

			/* Calculate the remaining volume */
			if (veItemIsValid(&tank->level) && veItemIsValid(&tank->capacity)) {
				VeVariant remaining, level, capacity;
				veItemValue(&tank->level, &level);
				veItemValue(&tank->capacity, &capacity);
				veVariantFloat(&remaining, level.value.Float * capacity.value.Float / 100);
				veItemOwnerSet(&tank->remaining, &remaining);
			}
			break;

		default:
			break;
		}
	}
}

veBool veCanTankAdd(VeSystem *system)
{
	VeVariant variant;
	VeTank *tank = (VeTank *) system->canbusCtx;

	system->master->hooks.rxFun = handleMessage;
	system->master->hooks.ctx = tank;

	tank->tankInstance = INVALID_INSTANCE;

	/* Product info */
	veItemOwnerSet(&system->product.productId, veVariantUn16(&variant, VE_PROD_ID_TANK_SENSOR));
	veItemOwnerSet(&system->product.productName, veVariantStr(&variant, veProductGetName(VE_PROD_ID_TANK_SENSOR)));

	msgRouterAddHook(&system->master->hooks);

	return veTrue;
}

/*
 * Static message handler that check if a device sends out a
 * fluid level PGN. If so, it is a tank sensor.
 */
static void handleMsg(void *ctx, struct J1939Func *func)
{
	VE_UNUSED(ctx);

	if (J1939MsgNrFromId(n2kMsgIn->msgId) != N2K_MSGNR_FLUID_LEVEL)
		return;

	func->isTank = veTrue;
	veCanCheckForNewDevice(func);
}

void veCanTankInit(void)
{
	hook.rxFun = handleMsg;
	msgRouterAddHook(&hook);
}
