#include <velib/velib_inc_J1939.h>

#include <msg_router.h>
#include <velib/io/ve_stream.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/J1939/J1939.h>
#include <velib/J1939/J1939_function.h>
#include <velib/nmea2k/n2k.h>
#include <velib/nmea2k/n2k_complex.h>
#include <velib/vecan/reg_rx.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_item_utils.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/products.h>

#include <values_vebus.h>
#include <vecan_common.h>
#include <vecan_vebus.h>

static VeVariant v;

static void veInDcStatusVI(VeItem *V, VeItem *I, VeItem *P)
{
	veInUn8(); /* instance */
	veInN2kDcV(V);
	veInN2kDcI(I);
	veItemCalculatePower(V, I, P);
}

static void veInJ1939VHzI(VeItem *V, VeItem *frequency, VeItem *current, VeItem *power)
{
	VeVariant v;

	veInUn16Le(); /* skip line / line voltage */
	veInJ1939Voltage(V);
	veInJ1939Hz(frequency);

	/*
	 * If power is given, the current is derived from it, to create a RMS Current
	 * containing a sign if power goes backwards.
	 */
	if (power == NULL) {
		veInJ1939Current(current);
	} else {
		if (veVariantIsValid(&power->variant) && veVariantIsValid(&V->variant) && V->variant.value.UN16 != 0) {
			veAssert(power->variant.type.tp == VE_SN32);
			veAssert(V->variant.type.tp == VE_UN16);
			veItemOwnerSet(current, veVariantFloat(&v, (float) power->variant.value.SN32 / V->variant.value.UN16));
		} else {
			veItemInvalidate(current);
		}
	}
}

veBool veCanVebusSetMode(VeSystem* system, int state)
{
	// Prepare a message.
	if (!veRegPrepareSf(system->master->nmt.nad, VE_REG_DEVICE_MODE))
		return veFalse;

	veOutUn8(state);
	veOutUn24Le(0);
	veOutSendSf();

	return veTrue;
}

/* Set AC2 current limit to limit in 0.1 amps */
static veBool vebusSetAcLimitAcIn(VeSystem* system, int n, un16 limit)
{
	if (!system->master)
		return veFalse;

	// Prepare a message.
	if (!veRegPrepareSf(system->master->nmt.nad,
				n == 1 ? VE_REG_AC_IN_1_CURRENT_LIMIT_REMOTE : VE_REG_AC_IN_2_CURRENT_LIMIT_REMOTE))
		return veFalse;

	veOutUn16Le(limit);
	veOutUn16Le(0);
	veOutSendSf();

	return veTrue;
}

/* Set AC1 current limit to limit in 0.1 amps */
veBool veCanVebusSetAcLimitAcIn1(VeSystem *system, un16 limit)
{
	return vebusSetAcLimitAcIn(system, 1, limit);
}

/* Set AC2 current limit to limit in 0.1 amps */
veBool veCanVebusSetAcLimitAcIn2(VeSystem *system, un16 limit)
{
	return vebusSetAcLimitAcIn(system, 2, limit);
}

/* limit in 0.1A */
veBool veCanVebusSetMaxChargeCurrent(VeSystem *system, un16 limit)
{
	if (!system->master)
		return veFalse;

	// Prepare a message.
	if (!veRegPrepareSf(system->master->nmt.nad, VE_REG_MK2CAN_MAX_CHARGE_CURRENT))
		return veFalse;

	veOutUn16Le(limit);
	veOutUn16Le(0);
	veOutSendSf();

	return veTrue;
}

void veInJ1939PowerAndApparent(struct VeItem *power, struct VeItem *apparent)
{
	veInJ1939Power(power);
	veInJ1939Power(apparent);
}

/* parse standard message from the inverter function */
void handleInverterMsg(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeBusSys* veSys = (VeBusSys*) ctx;
	VE_UNUSED(func);

	// by default message are only processed once and then freed.
	canId = n2kMsgIn->msgId;

	// Remove target address for device specific messages.
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
	{
		// handle incoming message for the first data page.
		switch(J1939PgnFromId(canId))
		{
		case J1939_PGN0_UTIL_UPAAC:
			veInJ1939VHzI(&veSys->items.VE_SYS_AC_OUT1_A_V,
							&veSys->items.VE_SYS_AC_OUT1_A_FREQUENCY,
							&veSys->items.VE_SYS_AC_OUT1_A_I,
							&veSys->items.VE_SYS_AC_OUT1_A_P);
			break;
		case J1939_PGN0_UTIL_UPAACP:
			veInJ1939PowerAndApparent(&veSys->items.VE_SYS_AC_OUT1_A_P, &veSys->items.VE_SYS_AC_OUT1_A_APPARENT_P);
			break;

		case J1939_PGN0_UTIL_UPBAC:
			veInJ1939VHzI(&veSys->items.VE_SYS_AC_OUT1_B_V,
							&veSys->items.VE_SYS_AC_OUT1_B_FREQUENCY,
							&veSys->items.VE_SYS_AC_OUT1_B_I,
							&veSys->items.VE_SYS_AC_OUT1_B_P);
			break;
		case J1939_PGN0_UTIL_UPBACP:
			veInJ1939PowerAndApparent(&veSys->items.VE_SYS_AC_OUT1_B_P, &veSys->items.VE_SYS_AC_OUT1_B_APPARENT_P);
			break;

		case J1939_PGN0_UTIL_UPCAC:
			veInJ1939VHzI(&veSys->items.VE_SYS_AC_OUT1_C_V,
							&veSys->items.VE_SYS_AC_OUT1_C_FREQUENCY,
							&veSys->items.VE_SYS_AC_OUT1_C_I,
							&veSys->items.VE_SYS_AC_OUT1_C_P);
			break;
		case J1939_PGN0_UTIL_UPCACP:
			veInJ1939PowerAndApparent(&veSys->items.VE_SYS_AC_OUT1_C_P, &veSys->items.VE_SYS_AC_OUT1_C_APPARENT_P);
			break;

		case J1939_PGN0_UTIL_UTACP:
			veInJ1939Power(&veSys->items.VE_SYS_AC_OUT1_P);
			break;

		default:
			;
		}
	} else {
		// handle incoming message for the first data page
		switch (J1939PgnFromId(canId))
		{
		case N2K_PGN1_DC_DETAILS:
			{
				un8 soc;

				veInUn8(); // SID
				veInUn8(); // instance
				veInUn8(); // DC type
				soc = veInUn8();
				if (validUN8(soc))
					veItemOwnerSet(&veSys->items.VE_VAR_SOC, veVariantFloat(&v, soc));
				else
					veItemInvalidate(&veSys->items.VE_VAR_SOC);
			}
			break;

		case N2K_PGN1_BATTERY_STATUS:
			veInDcStatusVI(&veSys->items.VE_SYS_UBAT, &veSys->items.VE_SYS_IBAT, &veSys->items.VE_SYS_DC_POWER);
			break;

		/*
		 * Firmware up to v0.11 wrongly use the switch control message.
		 * v0.12 and up use the switch status with is intended for binary events.
		 */
		case N2K_PGN1_SWITCH_CONTROL:
		case N2K_PGN1_SWITCH_STATUS:
			{
				un8 bankInstance = veInUn8();
				un8 n;
				VeItem *item;

				if (bankInstance == 0) {
					un8 warning;
					un8 alarm;

					/* Skip reserved inverter / charger on / off */
					veInBits8Le(2);
					veInBits8Le(2);
					veInBits8Le(2);
					veInBits8Le(2);

					warning = veInBits8Le(2);
					alarm = veInBits8Le(2);
					if (alarm == 1)
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_LOW_BATTERY, veVariantUn8(&v, 2));
					else
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_LOW_BATTERY, veVariantUn8(&v, warning == 1 ? 1 : 0));
					warning = veInBits8Le(2);
					alarm = veInBits8Le(2);
					if (alarm == 1)
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_TEMPERATURE, veVariantUn8(&v, 2));
					else
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_TEMPERATURE, veVariantUn8(&v, warning == 1 ? 1 : 0));
					warning = veInBits8Le(2);
					alarm = veInBits8Le(2);
					if (alarm == 1)
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_OVERLOAD, veVariantUn8(&v, 2));
					else
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_OVERLOAD, veVariantUn8(&v, warning == 1 ? 1 : 0));
					warning = veInBits8Le(2);
					alarm = veInBits8Le(2);
					if (alarm == 1)
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_U_BAT_RIPPLE, veVariantUn8(&v, 2));
					else
						veItemOwnerSet(&veSys->items.VE_SYS_ALARM_U_BAT_RIPPLE, veVariantUn8(&v, warning == 1 ? 1 : 0));

				} else if (bankInstance == 1) {

					/* update the led status */
					item = &veSys->items.VE_SYS_LED_MAINS;
					for (n = 0; n < 8; n++) {
						un8 on = veInBits8Le(2);
						un8 blink = veInBits8Le(2);
						un8 state;
						/* check for invalid state */
						if (on > 1 || blink > 1)
							state = 4;
						else
							state = 2 * blink + on;
						veItemOwnerSet(item, veVariantBit(&v, 2, state));
						item++;
					}
				}
			}
			break;

		default:
			;
		}
	}
}

/* parse standard message from the ac input function */
void handleAcInMsg(void *ctx, struct J1939Func *func)
{
	J1939Id canId;
	VeBusSys* veSys = (VeBusSys*) ctx;
	VE_UNUSED(func);

	// by default message are only processed once and then freed.
	canId = n2kMsgIn->msgId;

	// Remove target address for device specific messages.
	if (canId.p.pf <= 0xEF)
		canId.p.ps = 0;

	if (n2kMsgIn->msgId.p.pg.p.dp == 0)
	{
		// handle incoming message for the first data page.
		switch(J1939PgnFromId(canId))
		{
		case J1939_PGN0_UTIL_UPAAC:
			veInJ1939VHzI(&veSys->items.VE_SYS_AC_IN_A_V,
							&veSys->items.VE_SYS_AC_IN_A_FREQUENCY,
							&veSys->items.VE_SYS_AC_IN_A_I,
							&veSys->items.VE_SYS_AC_IN_A_P);
			break;
		case J1939_PGN0_UTIL_UPAACP:
			veInJ1939PowerAndApparent(&veSys->items.VE_SYS_AC_IN_A_P, &veSys->items.VE_SYS_AC_IN_A_APPARENT_P);
			break;

		case J1939_PGN0_UTIL_UPBAC:
			veInJ1939VHzI(&veSys->items.VE_SYS_AC_IN_B_V,
							&veSys->items.VE_SYS_AC_IN_B_FREQUENCY,
							&veSys->items.VE_SYS_AC_IN_B_I,
							&veSys->items.VE_SYS_AC_IN_B_P);
			if (veSys->numberOfPhases < 2) {
				veSys->numberOfPhases = 2;
				veItemOwnerSet(&veSys->items.VE_SYS_PHASES, veVariantUn8(&v, veSys->numberOfPhases));
			}
			break;
		case J1939_PGN0_UTIL_UPBACP:
			veInJ1939PowerAndApparent(&veSys->items.VE_SYS_AC_IN_B_P, &veSys->items.VE_SYS_AC_IN_B_APPARENT_P);
			break;

		case J1939_PGN0_UTIL_UPCAC:
			veInJ1939VHzI(&veSys->items.VE_SYS_AC_IN_C_V,
							&veSys->items.VE_SYS_AC_IN_C_FREQUENCY,
							&veSys->items.VE_SYS_AC_IN_C_I,
							&veSys->items.VE_SYS_AC_IN_C_P);
			if (veSys->numberOfPhases < 3) {
				veSys->numberOfPhases = 3;
				veItemOwnerSet(&veSys->items.VE_SYS_PHASES, veVariantUn8(&v, veSys->numberOfPhases));
			}
			break;
		case J1939_PGN0_UTIL_UPCACP:
			veInJ1939PowerAndApparent(&veSys->items.VE_SYS_AC_IN_C_P, &veSys->items.VE_SYS_AC_IN_C_APPARENT_P);
			break;

		case J1939_PGN0_UTIL_UTACP:
			veInJ1939Power(&veSys->items.VE_SYS_AC_IN_P);
			break;

		default:
			;
		}
	} else {
		// handle incoming message for the first data page
		switch (J1939PgnFromId(canId))
		{
		default:
			;
		}
	}
}

static void combineProductAndVersion(struct VeItem *productId, struct VeItem *versionPart, struct VeItem *combined)
{
	VeVariant v, w;

	if (!veItemIsValid(productId) || !veItemIsValid(versionPart)) {
		veItemInvalidate(combined);
	} else {
		veVariantUn32(&w,
							decimalToHex(veItemValue(productId, &v)->value.UN16) * 1000 +
							decimalToHex(veItemValue(versionPart, &v)->value.UN32)
						);
		veItemOwnerSet(combined, &w);
	}
}

static void productIdOrVersionChanged(VeBusSys* vebus, un8 id)
{
	switch (id) {
	case 1:
		combineProductAndVersion(&vebus->mk2ProductId, &vebus->mk2VersionPart, &vebus->items.VE_SYS_MK2_VERSION);
		break;
	case 2:
		combineProductAndVersion(&vebus->masterProductId, &vebus->masterVersionPart, &vebus->items.VE_SYS_VEBUS_VERSION);
		break;
	case 0x80:
		combineProductAndVersion(&vebus->dmcProductId, &vebus->dmcVersionPart, &vebus->items.VE_SYS_DMC_VERSION);
		break;
	case 0x81:
		combineProductAndVersion(&vebus->bmsProductId, &vebus->bmsVersionPart, &vebus->items.VE_SYS_BMS_VERSION);
		break;
	default:
		;
	}

	{
		// The vebus system can not be remotely controlled if either a vebus DMC or BMS is present
		int adjustable = !veItemIsValid(&vebus->items.VE_SYS_DMC_VERSION) && !veItemIsValid(&vebus->items.VE_SYS_BMS_VERSION);
		veItemOwnerSet(&vebus->items.VE_SYS_MODE_IS_ADJUSTABLE, veVariantBit(&v, 2, adjustable));
		veItemOwnerSet(&vebus->items.VE_SYS_CURRENT_LIMIT_IS_ADJUSTABLE, veVariantBit(&v, 2, adjustable));
	}
}

/*=== prorietary messages from the vebus system ===*/
void handleInverterVreg(void *ctx, struct J1939Func *src,
						VeEvent ev, VeRegId id, VeRegAckCode code)
{
	VeBusSys* vebus = (VeBusSys*) ctx;
	VE_UNUSED(src);
	VE_UNUSED(code);

	switch (ev)
	{
	case VE_EV_VREG_RX:
		switch ((int) id)
		{
		case VE_REG_APP_VER:
			{
				un8 identifier = veInUn8();
				VeVariant version;

				version.type.tp = VE_UN32;
				version.value.UN32 = veInUn24Le();
				if (version.value.UN32 == 0xFFFFFF)
					veVariantInvalidate(&version);

				if (identifier == 1)
					veItemOwnerSet(&vebus->mk2VersionPart, &version);
				else if (identifier == 2)
					veItemOwnerSet(&vebus->masterVersionPart, &version);
				else if (identifier == 0x80)
					veItemOwnerSet(&vebus->dmcVersionPart, &version);
				else if (identifier == 0x81)
					veItemOwnerSet(&vebus->bmsVersionPart, &version);
				productIdOrVersionChanged(vebus, identifier);
			}
			break;

		case VE_REG_PRODUCT_ID:
			{
				un8 identifier = veInUn8();
				un16 prodId = veInUn16Le();
				if (identifier == 1)
					veItemOwnerSet(&vebus->mk2ProductId, veVariantUn16(&v, prodId));
				else if (identifier == 2)
					veItemOwnerSet(&vebus->masterProductId, veVariantUn16(&v, prodId));
				else if (identifier == 0x80)
					veItemOwnerSet(&vebus->dmcProductId, veVariantUn16(&v, prodId));
				else if (identifier == 0x81)
					veItemOwnerSet(&vebus->bmsProductId, veVariantUn16(&v, prodId));
				productIdOrVersionChanged(vebus, identifier);
			}
			break;

		case VE_REG_DEVICE_MODE:
			veInItemLe(&vebus->items.VE_SYS_MODE, VE_UN8, 0);
			break;

		case VE_REG_DEVICE_STATE:
			veItemOwnerSet(&vebus->items.VE_SYS_STATE, veVariantUn8(&v, veInUn8()));
			break;

		case VE_REG_REMOTE_CONTROL_USED:
			{
				/*
				 * Some proprietary features must be enabled by a panel, before
				 * they get active, some examples are:
				 *  - sending leds status to imitate a PMC
				 *  - overwrite current limiters by a remote panels.
				 * The enabling is just a matter of sending a vreg command for
				 * this register with the supported features set. The remote side
				 * will OR so all features of alll connected panels are enabled.
				 * A check is made that the needed feature are (still), if not
				 * they are enabled.
				 * Note: mask the value, more features can be added and might be
				 * set by other devices communicating with the device.
				 */
				un32 remoteControl = veInUn32Le();
				if ((remoteControl & vebus->can.remoteControl) != vebus->can.remoteControl)
					vebus->can.sendEnableFeatures = veTrue;
			}
			break;

		case VE_REG_AC_IN_CURRENT_LIMIT:
			veInItemLe(&vebus->items.VE_SYS_AC_IN_LIM, VE_UN16, 1);
			break;

		case VE_REG_AC_IN_ACTIVE:
			veInItemLe(&vebus->items.VE_SYS_AC_IN_ACTIVE, VE_UN8, 0);
			break;

		case VE_REG_MK2CAN_MAX_CHARGE_CURRENT:
			veInItemLe(&vebus->items.VE_I_BAT_BULK, VE_UN16, 1);
			break;

		case VE_REG_MK2CAN_VEBUS_ERROR:
			veItemOwnerSet(&vebus->items.VE_SYS_VEBUS_ERROR, veVariantUn32(&v, veInUn8()));
			break;

		/* kWh counters */
		case VE_REG_ENERGY_FROM_ACIN1_TO_INVERTER:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_AC_IN1_TO_INVERTER);
			break;

		case VE_REG_ENERGY_FROM_ACIN2_TO_INVERTER:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_AC_IN2_TO_INVERTER);
			break;

		case VE_REG_ENERGY_FROM_ACIN1_TO_ACOUT:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_AC_IN1_TO_AC_OUT);
			break;

		case VE_REG_ENERGY_FROM_ACIN2_TO_ACOUT:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_AC_IN2_TO_AC_OUT);
			break;

		case VE_REG_ENERGY_FROM_INVERTER_TO_ACIN1:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_INVERTER_TO_AC_IN1);
			break;

		case VE_REG_ENERGY_FROM_INVERTER_TO_ACIN2:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_INVERTER_TO_AC_IN2);
			break;

		case VE_REG_ENERGY_FROM_ACOUT_TO_ACIN1:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_AC_OUT_TO_AC_IN1);
			break;

		case VE_REG_ENERGY_FROM_ACOUT_TO_ACIN2:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_AC_OUT_TO_AC_IN2);
			break;

		case VE_REG_ENERGY_FROM_INVERTER_TO_ACOUT:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_INVERTER_TO_AC_OUT);
			break;

		case VE_REG_ENERGY_FROM_ACOUT_TO_INVERTER:
			veInN2kEnergy(&vebus->items.VE_SYS_ENERGY_OUT_TO_INVERTER);
			break;

		/* Hub1 voltage setpoint */
		case VE_REG_LINK_VSET:
			// NOTE: this one is unsigned unlike regular voltages!
			veInItemLe(&vebus->hub1VoltageSetPoint, VE_UN16, 2);
			break;

		default:
			;
		}

	case VE_EV_VREG_RX_ACK:
		break;

	default:
		;
	}
}

static void update(VeSystem *system)
{
	VeBusSys *vebus = (VeBusSys *) system->canbusCtx;

	if (vebus->numberOfPhases != 1) {
		if (system->master->nmt.flags & J1939_NMT_DEV_DISCONNECTED) {
			vebus->numberOfPhases = 1;
			veItemOwnerSet(&vebus->items.VE_SYS_PHASES, veVariantUn8(&v, vebus->numberOfPhases));
		}
	}

	if (vebus->can.sendEnableFeatures) {
		if (veRegPrepareSfWhenOnBus(system->master, VE_REG_REMOTE_CONTROL_USED)) {
			vebus->can.sendEnableFeatures = veFalse;
			veOutUn32Le(vebus->can.remoteControl);
			veOutSendSf();
		}
	}

	if (vebus->can.reqVregs)
		if (veRegRequestMaskWhenOnBus(system->master, VE_REG_DEVICE_MODE, 0xFF00))
			vebus->can.reqVregs = veFalse;

	if (vebus->can.getProductInfo) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_PRODUCT_ID) &&
				veRegRequestWhenOnBus(system->master, VE_REG_APP_VER)) {
			vebus->can.getProductInfo = veFalse;
		}
	}

	if (vebus->can.getVeBusError) {
		if (veRegRequestWhenOnBus(system->master, VE_REG_MK2CAN_VEBUS_ERROR))
			vebus->can.getVeBusError = veFalse;
	}

	if (vebus->can.getDcDetails) {
		if (j1939RequestPgn(system->master->nmt.nad, 1, N2K_PGN1_DC_DETAILS))
			vebus->can.getDcDetails = veFalse;
	}
}

static void tick(VeSystem *system)
{
	VeBusSys *vebus = (VeBusSys *) system->canbusCtx;

	if (vebus->can.pollTimer == 0 || --vebus->can.pollTimer == 0) {
		vebus->can.getProductInfo = veTrue;
		vebus->can.getVeBusError = veTrue;
		vebus->can.getDcDetails = veTrue;
		vebus->can.pollTimer = 6 /* sec */ * 20;
	}
}

void veCanVebusFindAcInPut(VeSystem *system)
{
	VeBusSys *vebus = (VeBusSys *) system->canbusCtx;
	J1939FuncTp *itt;

	if (vebus->acInput != NULL)
		return;

	itt = j1939FuncList;

	while (itt)
	{
		un32 id = system->master->nmt.name.low + 1;

		if (itt->nmt.name.low == id)
		{
			vebus->acInput = itt;
			vebus->acInput->hooks.func = itt;
			vebus->acInput->hooks.rxFun = handleAcInMsg;
			vebus->acInput->hooks.vregFun = NULL;
			vebus->acInput->hooks.ctx = system->canbusCtx;
			msgRouterAddHook(&vebus->acInput->hooks);

			return;
		}
		itt = itt->next;
	}
}

veBool veCanVebusAdd(VeSystem *system)
{
	VeBusSys *vebus = (VeBusSys *) system->canbusCtx;

	/* Send our capibabilities to the CAN interface */
	vebus->can.remoteControl = VE_REG_RC_ON_OFF_MODE | VE_REG_RC_SEND_LEDS | VE_REG_RC_AC2_IN_LIMIT | VE_REG_RC_AC1_IN_LIMIT;
	vebus->can.sendEnableFeatures = veTrue;
	vebus->can.reqVregs = veTrue;
	vebus->can.pollTimer = 0;
	vebus->can.getProductInfo = veTrue;
	vebus->can.getVeBusError = veTrue;
	vebus->can.getDcDetails = veTrue;

	// Workaround: numberOfPhases depends on if J1939_PGN0_UTIL_UPBAC or J1939_PGN0_UTIL_UPCAC is received.
	vebus->numberOfPhases = 1;
	veItemOwnerSet(&vebus->items.VE_SYS_PHASES, veVariantUn8(&v, vebus->numberOfPhases));

	system->update = update;
	system->tick = tick;
	system->master->hooks.func = system->master;
	system->master->hooks.ctx = system->canbusCtx;
	system->master->hooks.rxFun = handleInverterMsg;
	system->master->hooks.vregFun = handleInverterVreg;
	msgRouterAddHook(&system->master->hooks);

	veCanVebusFindAcInPut(system);

	return veTrue;
}
