#include <stdlib.h>

#include <velib/base/base.h>
#include <velib/base/ve_string.h>
#include <velib/types/ve_item.h>

#include <dbus_service.h>
#include <values_battery.h>
#include <values_common.h>
#include <vecan_battery.h>

/**
 * Definitions how to make a textual representation of a value.
 * @see veVariantFmt in types/ve_variant_print.c.
 */
static VeVariantUnitFmt dcV = {2, "V"};
static VeVariantUnitFmt dcI = {1, "A"};
static VeVariantUnitFmt dcW = {0, "W"};
static VeVariantUnitFmt ampH = {1, "Ah"};
static VeVariantUnitFmt kwh = {2, "kWh"};
static VeVariantUnitFmt sec = {0, "s"};
static VeVariantUnitFmt none = {0, ""};
static VeVariantUnitFmt percent = {1, "%"};

/**
 * The tree of known values of a BMV.
 */

#define offset(a)	offsetof(VeBattery, a)

static ItemInfo itemInfo[] = {
	{"Dc/0/Voltage",									&dcV,	offset(dcStatusBattery.voltage),	8	},
	{"Dc/0/Current",									&dcI,	offset(dcStatusBattery.current),	8	},
	{"Dc/0/Power",										&dcW,	offset(dcStatusBattery.power),		8	},
	{"Dc/1/Voltage",									&dcV,	offset(dcStatusAux.voltage),		8	},
	{"ConsumedAmphours",								&ampH,	offset(consumedAmphours),			8	},
	{"Soc",												&percent,	offset(stateOfCharge),			8	},
	{"TimeToGo",										&sec,	offset(timeToGo),					8	},
	{"Alarms/Alarm",									&none,	offset(alarms.alarm),				8	},
	{"Alarms/LowVoltage",								&none,	offset(alarms.lowVoltage),			8	},
	{"Alarms/HighVoltage",								&none,	offset(alarms.highVoltage),			8	},
	{"Alarms/LowStarterVoltage",						&none,	offset(alarms.lowStarterVoltage),	8	},
	{"Alarms/HighStarterVoltage",						&none,	offset(alarms.highStarterVoltage),	8	},
	{"Alarms/LowSoc",									&none,	offset(alarms.lowSoc),				8	},
	{"Relay/0/State",									&none,	offset(alarms.relayState),			8	},
	{"History/DeepestDischarge",						&ampH,	offset(history.h[0]),				8	},
	{"History/LastDischarge",							&ampH,	offset(history.h[1]),				8	},
	{"History/AverageDischarge",						&ampH,	offset(history.h[2]),				8	},
	{"History/ChargeCycles",							&none,	offset(history.h[3]),				8	},
	{"History/FullDischarges",							&none,	offset(history.h[4]),				8	},
	{"History/TotalAhDrawn",							&ampH,	offset(history.h[5]),				8	},
	{"History/MinimumVoltage",							&dcV,	offset(history.h[6]),				8	},
	{"History/MaximumVoltage",							&dcV,	offset(history.h[7]),				8	},
	{"History/TimeSinceLastFullCharge",					&sec,	offset(history.h[8]),				8	},
	{"History/AutomaticSyncs",							&none,	offset(history.h[9]),				8	},
	{"History/LowVoltageAlarms",						&none,	offset(history.h[10]),				8	},
	{"History/HighVoltageAlarms",						&none,	offset(history.h[11]),				8	},
	{"History/LowStarterVoltageAlarms",					&none,	offset(history.h[12]),				8	},
	{"History/HighStarterVoltageAlarms",				&none,	offset(history.h[13]),				8	},
	{"History/MinimumStarterVoltage",					&dcV,	offset(history.h[14]),				8	},
	{"History/MaximumStarterVoltage",					&dcV,	offset(history.h[15]),				8	},
	{"History/DischargedEnergy",						&kwh,	offset(history.h[16]),				8	},
	{"History/ChargedEnergy",							&kwh,	offset(history.h[17]),				8	},
	{"Settings/HasStarterVoltage",						&none,	offset(hasStarterVoltage),			8	},
	{"Settings/HasMidVoltage",							&none,	offset(hasMidVoltage),				8	},
	{"Settings/HasTemperature",							&none,	offset(hasTemperature),				8	},
	{"Settings/HasSettings",							&none,	offset(hasSettings),				8	},
	{"Settings/DischargeFloorLinkedToRelay",			&none,	offset(dischargeFloorLinkedToRelay),8	},
};

#undef offset

/*
 * Create a BMV system object with the corresponding item tree.
 *
 * The tree with items can be used to navigate through the values,
 * format / set / get them etc.
 */
veBool valuesBmvInit(VeSystem *system)
{
	VeBattery *battery = calloc(1, sizeof(VeBattery));

	if (!battery)
		return veFalse;

	veVariantUn32(&battery->hasTemperature.variant, 0);
	veVariantUn32(&battery->hasStarterVoltage.variant, 1);
	veVariantUn32(&battery->hasMidVoltage.variant, 0);

	/* No settings support (yet) */
	battery->minSettingsVersion = 0xFFFF;

	/* In the BMV the relay and discharge floor are still linked */
	battery->minRelayLowSocVersion = 0xFFFF;

	/* create the values tree to store the variables */
	system->canbusCtx = battery;
	valuesInitTree(itemInfo, ARRAY_LENGTH(itemInfo), &system->values, system);

	/* setup the CAN bus to fill them in */
	if (!veCanBatteryAdd(system)) {
		free(battery);
		return veFalse;
	}

	return veTrue;
}
