#include <stdlib.h>

#include <velib/base/base.h>
#include <velib/types/variant_print.h>
#include <velib/types/ve_item.h>
#include <velib/base/ve_string.h>

#include <values_common.h>
#include <values_charger.h>
#include <vecan_charger.h>

/**
 * Definitions how to make a textual representation of a value.
 * @see veVariantFmt in types/ve_variant_print.c.
 */
static VeVariantUnitFmt dcV = {2, "V"};
static VeVariantUnitFmt dcI = {1, "A"};
static VeVariantUnitFmt acI = {1, "A"};
static VeVariantUnitFmt dcT = {0, " degrees"};
static VeVariantUnitFmt kWh = {2, "kWh"};
static VeVariantUnitFmt W = {0, "W"};
static VeVariantUnitFmt none = {0, ""};

#define offset(a)	offsetof(VeCharger, a)

/**
 * The common values of a charger system.
 * @note the items typically contain SI units, with floats if needed.
 */
static ItemInfo commonItemInfo[] = {
	{"State",						&none,	offset(state),						8},
	{"Mode",						&none,	offset(mode),						8},
	{"ErrorCode",					&none,	offset(errorCode),					8},
	{"Relay/0/State",				&none,	offset(alarms.relayState),			8},
	{"Alarms/LowVoltage",			&none,	offset(alarms.lowVoltage),			8},
	{"Alarms/HighVoltage",			&none,	offset(alarms.highVoltage),			8},
	{"NrOfOutputs",					&none,	offset(outputs),					0}
};

static ItemInfo solarItemInfo[] = {
	{"Dc/0/Voltage",				&dcV,	offset(dcStatus[0].voltage),		8},
	{"Dc/0/Current",				&dcI,	offset(dcStatus[0].current),		8},
	{"Dc/0/Temperature",			&dcT,	offset(temperature),				8},
	{"Pv/V",						&dcV,	offset(pvStatus.voltage),			8},
	{"Pv/I",						&dcI,	offset(pvStatus.current),			8},
	{"Yield/System",				&kWh,	offset(yield.system),				0},
	{"Yield/User",					&kWh,	offset(yield.user),					0},
	{"Yield/Power",					&W,		offset(yield.power),				0},
	{"History/Daily/0/Yield",		&kWh,	offset(hist[0].yield),				8},
	{"History/Daily/0/MaxPower",	&W,		offset(hist[0].maxPower),			8},
	{"History/Daily/1/Yield",		&kWh,	offset(hist[1].yield),				8},
	{"History/Daily/1/MaxPower",	&W,		offset(hist[1].maxPower),			8}
};

static ItemInfo acItemInfo[] = {
	{"Dc/0/Voltage",				&dcV,	offset(dcStatus[0].voltage),		8},
	{"Dc/0/Current",				&dcI,	offset(dcStatus[0].current),		8},
	{"Dc/0/Temperature",			&dcT,	offset(temperature),				8},
	{"Dc/1/Voltage",				&dcV,	offset(dcStatus[1].voltage),		8},
	{"Dc/1/Current",				&dcI,	offset(dcStatus[1].current),		8},
	{"Dc/2/Voltage",				&dcV,	offset(dcStatus[2].voltage),		8},
	{"Dc/2/Current",				&dcI,	offset(dcStatus[2].current),		8},
	{"Ac/In/L1/I",					&acI,	offset(acIn.current),				8},
	{"Ac/In/L1/P",					&W,		offset(acIn.power),					8},
	{"Ac/In/CurrentLimit",			&acI,	offset(acInCurrentLimit),			8}
};

#undef offset

/* stub to forward state change request. */
static veBool setMode(VeItem *item, void *ctx, VeVariant *variant)
{
	VE_UNUSED(item);

	return veCanChargerSetMode((VeSystem *)ctx, variant->value.SN32);
}

/* stub to forward state ac current limit */
static veBool setCurrentLimit(VeItem *item, void *ctx, VeVariant *variant)
{
	VE_UNUSED(item);

	veVariantToFloat(variant);
	variant->value.Float *= 10;
	veVariantFromFloat(variant, VE_UN16);

	return veCanChargerSetCurrentLimit((VeSystem *)ctx, variant->value.UN16);
}

static veBool init(VeSystem *system, ItemInfo const *extItemInfo, size_t count)
{
	VeCharger *charger = calloc(1, sizeof(VeCharger));

	if (!charger)
		return veFalse;

	/* create the values tree to store the variables */
	system->canbusCtx = charger;

	valuesInitTree(commonItemInfo, ARRAY_LENGTH(commonItemInfo), &system->values, system);

	if (extItemInfo)
		valuesInitTree(extItemInfo, count, &system->values, system);

	/* setup the CAN bus to fill them in */
	if (!veCanChargerAdd(system)) {
		free(charger);
		return veFalse;
	}

	veItemSetSetter(&charger->mode, setMode, system);

	return veTrue;
}

/*
 * Create a dbus system object and the corresponding item tree.
 *
 * The tree with items can be used to navigate through the values in
 * dbus system, format / set / get them etc.
 */
veBool valuesChargerInit(VeSystem *system)
{
	return init(system, NULL, 0);
}

veBool valuesSolarChargerInit(VeSystem *system)
{
	return init(system, solarItemInfo, ARRAY_LENGTH(solarItemInfo));
}

veBool valuesAcChargerInit(VeSystem *system)
{
	if (init(system, acItemInfo, ARRAY_LENGTH(acItemInfo))) {
		VeCharger *charger = (VeCharger *)system->canbusCtx;

		veItemSetSetter(&charger->acInCurrentLimit, setCurrentLimit, system);

		return veTrue;
	}

	return veFalse;
}
