#include <stdlib.h>

#include <velib/velib_inc_J1939.h>

#include <velib/base/ve_string.h>
#include <velib/types/ve_str.h>
#include <velib/types/ve_values.h>
#include <velib/types/ve_item.h>
#include <velib/types/ve_dbus_item.h>
#include <velib/types/ve_str.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_item_utils.h>
#include <velib/vecan/products.h>
#include <velib/canhw/canhw_driver.h>
#include <velib/platform/plt.h>
#include <velib/platform/console.h>

#include <dbus_service.h>
#include <senders.h>
#include <values_charger.h>
#include <values_common.h>
#include <values_vebus.h>
#include <values_battery.h>
#include <values_lynxion.h>
#include <vecan_charger.h>
#include <vecan_common.h>
#include <vecan_vebus.h>
#include <vecan_tank.h>

#define MODULE	"vecan"

#define setChildCtx(item) sysItt->item.ctx.ptr = &sysItt->child->item

/** The systems found on the CAN bus */
static VeSystem *systems;

/**
 * The root of the tree to locate values and format them etc.
 * This contains all the variables of all the system and uses
 * strings to look them up.
 */
static VeItem *itemTree;
static VeItem *devicesItem;
static VeItem *settingsItem;

static VeVariantUnitFmt none = {0, ""};

size_t versionFmt(VeVariant *var, void const *ctx, char* buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var))
		return ve_snprintf(buf, len, "%s", "");

	if (var->value.UN32 & 0x00FF0000) {
		return ve_snprintf(buf, len, "v%X.%02X.%02X",
				(var->value.UN32 >> 16) & 255, (var->value.UN32 >> 8) & 255, var->value.UN32 & 255);
	}
	return ve_snprintf(buf, len, "v%X.%02X", (var->value.UN32 >> 8) & 255, var->value.UN32 & 255);
}

size_t pidFmt(VeVariant *var, void const *ctx, char* buf, size_t len)
{
	VE_UNUSED(ctx);

	if (!veVariantIsValid(var))
		return ve_snprintf(buf, len, "%s", "");
	return ve_snprintf(buf, len, "0x%X", var->value.UN16);
}

/** 50ms tick to invalidate items if there are not received in time */
void valuesUpdate(void)
{
	VeSystem *system = systems;

	while (system)
	{
		if (system->update)
			system->update(system);
		system = system->next;
	}
}

/** used by hub1, forward the voltage setpoint from the multi to the CAN-bus */
static veBool forwardLinkVoltage(struct VeItem *item, void *ctx, VeVariant *variant)
{
	static VLinkDevice *vLink;
	VE_UNUSED(ctx);

	if (vLink == NULL) {
		VeVariant defaultValue, minValue, maxValue;
		VeItem *nad, *instance;

		nad = veItemGetOrCreateUid(settingsItem, "Multi/0/Nad");
		instance = veItemGetOrCreateUid(settingsItem, "Multi/0/Instance");

		if (!veDBusAddLocalSetting(nad,
								   veVariantUn8(&defaultValue, 0x8),
								   veVariantUn8(&minValue, 0),
								   veVariantUn8(&maxValue, J1939_NAD_MAX),
								   veFalse)) {
			logE("task", "veDBusAddLocalSetting failed");
			pltExit(10);
		}

		if (!veDBusAddLocalSetting(instance,
								   veVariantUn8(&defaultValue, 0x0),
								   veVariantUn8(&minValue, 0),
								   veVariantUn8(&maxValue, 0xFE),
								   veFalse)) {
			logE("task", "veDBusAddLocalSetting failed");
			pltExit(10);
		}

		vLink = vLinkNewInstance(1, instance, nad, item);
	}

	return veItemOwnerSet(item, variant);
}

void veCanServiceInit(struct VeItem **mainNad)
{
	VeStr s;
	char canId[100];
	VeItem *linkVoltageSetpoint;
	VeItem *mainService;
	VeItem *settingsService;
	VeVariant v;

	pltCanGwId(canId, sizeof(canId));

	veStrNewFormat(&s, "com.victronenergy.vecan.%s", canId);
	mainService = veItemGetOrCreateUid(itemTree, veStrCStr(&s));
	veStrFree(&s);

	devicesItem = veItemGetOrCreateUid(mainService, "Devices");

	/**
	 * This is a special item, when data is written to it, it will create a device
	 * which will announce itself as a multi and broadcast given setpoint. Solar
	 * chargers will pick this up and start behaving as a hub1 system.
	 * The setpoint needs to updated regularly or will become invalid (and hence the
	 * solar chargers are going to do their own thing again)
	 *
	 * NOTE: The CAN-bus can support multi hub1s. Only one is supported here for now.
	 */
	linkVoltageSetpoint = veItemCreateQuantity(mainService, "Link/ChargeVoltage", veVariantInvalidType(&v, VE_FLOAT), &veUnitVolt2Dec);
	veItemSetSetter(linkVoltageSetpoint, forwardLinkVoltage, NULL);
	veItemSetTimeout(linkVoltageSetpoint, 60);

	{
		struct VeDbus *dbus = exportToDbus(veItemId(mainService), mainService);
		VeVariant defaultValue, minValue, maxValue;

		if (dbus == NULL) {
			logE(MODULE, "export to dbus failed -> exit");
			exit(2);
		}

		/** import settings from the dbus */
		veDbusSetListeningDbus(dbus);

		settingsService = veItemGetOrCreateUid(itemTree, "com.victronenergy.settings");
		if (!veDbusAddRemoteService("com.victronenergy.settings", settingsService, veTrue)) {
			logE("task", "veDbusAddRemoteService failed, localsettings not running?");
			pltExit(1);
		}

		veStrNewFormat(&s, "Settings/Vecan/%s", canId);
		settingsItem = veItemGetOrCreateUid(settingsService, veStrCStr(&s));
		veStrFree(&s);

		*mainNad = veItemGetOrCreateUid(settingsItem, "MainInterface/Nad");

		if (!veDBusAddLocalSetting(*mainNad,
								   veVariantUn8(&defaultValue, 0xF0),
								   veVariantUn8(&minValue, 0),
								   veVariantUn8(&maxValue, J1939_NAD_MAX),
								   veFalse)) {
			logE("task", "veDBusAddLocalSetting failed");
			pltExit(10);
		}
	}
}

void valuesInit(VeItem **mainNad)
{
	itemTree = veValueTree();
	veCanChargersInit();
	veCanTankInit();
	veCanServiceInit(mainNad);
}

/** 50ms tick to invalidate items if there are not received in time */
void valuesTick(void)
{
	VeSystem *system = systems;

	while (system)
	{
		if (system->tick)
			system->tick(system);
		system = system->next;
	}

	veItemTick(itemTree);
}

#define itemFromOffset(src, offset) \
	(VeItem *)((un8 *)src + offset)

VeSetting *valuesGetSettingByItem(VeSettingCtx *ctx, VeItem *needle)
{
	VeSetting *setting = ctx->settings;

	while (setting && setting->id != VE_REG_INVALID) {
		VeItem *item = itemFromOffset(ctx->system->canbusCtx, setting->offset);
		if (item == needle)
			return setting;
		setting++;
	}

	return NULL;
}

static veBool setSetting(VeItem *item, void *ctx, VeVariant *variant)
{
	VeSettingCtx *settingCtx = (VeSettingCtx *)ctx;

	VeSetting *setting;
	if ((setting = valuesGetSettingByItem(settingCtx, item)))
		return veCanSetVeReg(settingCtx->system, variant, setting);

	return veFalse;
}

static void setSettingSetters(VeSystem *system)
{
	VeSetting *setting = (VeSetting *)system->settingsCtx.settings;

	while (setting && setting->id != VE_REG_INVALID) {
		VeItem *item = itemFromOffset(system->canbusCtx, setting->offset);

		VeVariant v;
		veItemSetMin(item, veVariantFloat(&v, setting->range->min));
		veItemSetMax(item, veVariantFloat(&v, setting->range->max));
		veItemSetDefault(item, veVariantFloat(&v, setting->range->def));

		veItemSetSetter(item, setSetting, &system->settingsCtx);
		setting++;
	}
}

/* Helper to create value trees */
void valuesInitTree(ItemInfo const *itemInfo, size_t count, VeItem *root, VeSystem *system)
{
	size_t n;

	for (n = 0; n < count; n++)
	{
		VeItem *item = (VeItem *) ((un8*)system->canbusCtx + itemInfo[n].offset);

		veItemAddChildByUid(root, itemInfo[n].id, item);
		veItemSetFmt(item, veVariantFmt, itemInfo[n].fmt);
		if (itemInfo[n].timeout != 0)
			veItemSetTimeout(item,  itemInfo[n].timeout);
	}

	/* If system contains settings, set their setters */
	setSettingSetters(system);
}

void valuesAddItem(VeItem *root, char const *id, VeItem *item, VeVariantUnitFmt *fmt, un8 timeout)
{
	/* Not a dbus item */
	if (id == NULL)
		return;

	/* Check if exists */
	if (veItemByUid(root, id))
		return;

	/* Add */
	veItemAddChildByUid(root, id, item);
	veItemSetFmt(item, veVariantFmt, fmt);
	if (timeout != 0)
		veItemSetTimeout(item, timeout);

	logI(MODULE, "added item %s", id);
}

/** fill buf with the gateway id as suitable for the dbus */
static veBool getGatewayIdForDbus(char *buf, size_t len)
{
	VeCanGateway *gw;
	size_t n;
	char *p;

	gw = veCanGwActive();
	n = veCanGwId(gw, buf, len);
	if (n >= len) {
		logE("values", "interface name too long!");
		return veFalse;
	}

	/*
	 * The gateway id is normally formatted as driver:interface, but the
	 * dbus can't cope with colons in the service name. Replace it with an
	 * underscore.
	 */
	p = buf;
	while (*p) {
		if (*p == ':')
			*p = '_';
		p++;
	}

	return veTrue;
}

/**
 * The original function, changes the device instance of the master only.
 * This is aviable inside the service itself.
 */
static veBool veCanSetDeviceInstanceMasterOfSytemRequest(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem *)ctx;
	VE_UNUSED(item);

	return veCanSetDeviceInstance(system->master, variant->value.UN8);
}

/**
 * Change the device instance of any device on the CAN-bus.
 * These are available in com.victronenergy.vebus
 */
static veBool veCanSetDeviceInstanceRequest(struct VeItem *item, void *ctx, VeVariant *variant)
{
	J1939FuncTp *dev = (J1939FuncTp *)ctx;
	VE_UNUSED(item);

	return veCanSetDeviceInstance(dev, variant->value.UN8);
}

void veCanCheckForNewDevice(struct J1939Func *dev)
{
	VeSystem *system, *sysItt;
	J1939NameHighFields *fields = &dev->nmt.name.high.fields;
	VeStr s;
	char buf[100];

	/* Dont parse the data from the device this program is sending towards the CAN-bus. */
	if (veCanIsLocalSender(dev))
		return;

	if (J1939_ACL_L_GET_MAN(dev->nmt.name.low) != N2K_MAN_VICTRON || fields->indGroup != J1939_INDUSTRY_MARINE ||
		!(fields->devClass == N2K_CLASS_ELECTRICAL_GENERATION || fields->devClass == N2K_CLASS_GENERAL)) { // lynx ion has class general (v1.19 and older)

		/* Offshore Systems Ltd tank sender adaptor */
		if (!(fields->devClass == N2K_CLASS_GENERAL && fields->function == N2K_FUNCTION_TRANSDUCER && dev->isTank))
			return;
	}

	switch (fields->function)
	{
	case N2K_FUNCTION_CHARGER:
		/* for chargers, only the master of a known group creates a system */
		if (!dev->isMaster || (dev->udf.flags & VE_DEV_GROUP_VALID) == 0)
			return;
		break;

	case N2K_FUNCTION_BATTERY:
		if (!dev->needsProductId) {
			dev->needsProductId = veTrue;
			return;
		}
		break;

	default:
		;
	}

	system = systems;
	while (system)
	{
		veAssert(system->master != NULL);

		if (system->master == dev) {

			/* Allow chargers, as its master could have become the master of another system */
			if (system->function != N2K_FUNCTION_CHARGER)
				return;

			/* Check if master has move to another device instance */
			if (system->instance ==	N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high))
				return;
		}

		switch (fields->function)
		{
		case N2K_FUNCTION_AC_INPUT:
			if (system->function == N2K_FUNCTION_INVERTER)
				veCanVebusFindAcInPut(system);
			break;

		case N2K_FUNCTION_CHARGER:

			if (system->function == N2K_FUNCTION_CHARGER) {

				/* Removed empty systems */
				if (veCanCountDevices(system) == 0) {
					logI(system->id, "switching master: %u -> ?", system->master->nmt.nad);
					logI(system->id, "empty system");
					system = veCanRemoveCharger(system);
					continue;
				}

				/* No switching if we have multiple masters.
				 * This is a configuration error and needs
				 * to be resolved by the user. */
				if (veCanCountMasters(system) > 1)
					return;

				/* Check switching of master */
				if (veCanChargerCheckMaster(system, dev))
					return;
			}
			break;

		default:
			;
		}
		system = system->next;
	}

	if (fields->function == N2K_FUNCTION_AC_INPUT)
		return;

	if (system)
		return;

	system = calloc(1, sizeof(VeSystem));
	if (!system)
		return;

	system->master = dev;
	dev->hooks.func = dev;
	dev->system = system;

	if (!getGatewayIdForDbus(buf, sizeof(buf)))
		goto error;

	if (fields->devClass == N2K_CLASS_ELECTRICAL_GENERATION) {
		switch(fields->function)
		{
		case N2K_FUNCTION_BATTERY:
		{
			VeDevVup* vup = NULL;
			VeDevVup* vupItt;
			un8 instance = 0;
			// Get the last vup instance.
			while ((vupItt = veDevUdfFindVup(dev, instance)) != NULL) {
				vup = vupItt;
				instance++;
			}

			if (!vup)
				goto error;

			switch (vup->prodId)
			{
			case VE_PROD_ID_VECAN_LYNX_ION:
			case VE_PROD_ID_VECAN_LYNX_ION_SHUNT:
				veStrNewFormat(&s, "com.victronenergy.bms.%s_di%d_uc%d", buf,
								N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
								J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
				system->id = s.data;
				if (!valuesLynxIonInit(system))
					goto error;

				if (vup->prodId == VE_PROD_ID_VECAN_LYNX_ION_SHUNT) {
					VeSystem *child = calloc(1, sizeof(VeSystem));
					VeStr str;
					if (!child)
						goto error;
					veStrNewFormat(&str, "com.victronenergy.battery.%s_di%d_uc%d", buf,
								N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
								J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
					child->id = str.data;
					child->parent = system;
					child->master = dev;
					system->child = child;

					if (!valuesLynxIonBatteryInit(child))
						goto error;
				}
				break;

			case VE_PROD_ID_VECAN_LYNX_SHUNT:
					veStrNewFormat(&s, "com.victronenergy.battery.%s_di%d_uc%d", buf,
								N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
								J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
				system->id = s.data;
				if (!valuesLynxShuntInit(system))
					goto error;
				break;

			default:
				veStrNewFormat(&s, "com.victronenergy.battery.%s_di%d_uc%d", buf,
								N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
								J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
				system->id = s.data;
				if (!valuesBmvInit(system))
					goto error;
			}
			break;
		}

		case N2K_FUNCTION_INVERTER:
			if (!valuesVebusInit(system))
				goto error;
			veStrNewFormat(&s, "com.victronenergy.vebus.%s_di%d_uc%d", buf,
							N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
							J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
			system->id = s.data;
			break;

#ifdef CHARGER
		case N2K_FUNCTION_CHARGER:
			switch (dev->udf.group)
			{
			case VE_GROUP_SOLAR_CHARGERS:
				if (!valuesSolarChargerInit(system))
					goto error;
				veStrNewFormat(&s, "com.victronenergy.solarcharger.%s_di%d_uc%d", buf,
								N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
								J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
				system->id = s.data;
				break;
			case VE_GROUP_GRID_CHARGERS:
				if (!valuesAcChargerInit(system))
					goto error;
				veStrNewFormat(&s, "com.victronenergy.charger.%s_di%d_uc%d", buf,
								N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
								J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
				system->id = s.data;
				break;
			default:
				if (fields->devClass != N2K_CLASS_GENERAL) { // workaround lynx ion has function charger.

					if (!valuesChargerInit(system))
						goto error;

					veStrNewFormat(&s, "com.victronenergy.charger.%s_di%d_uc%d", buf,
									N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
									J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
					system->id = s.data;
				}
				else {
				}
			}
			break;
#endif

		default:
			goto error;
		}
	}
	else if (fields->devClass == N2K_CLASS_GENERAL) {
		switch(fields->function)
		{
		case N2K_FUNCTION_INSTRUMENT:

			if (!valuesLynxIonInit(system))
				goto error;

			veStrNewFormat(&s, "com.victronenergy.bms.%s_di%d_uc%d", buf,
							N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
							J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
			system->id = s.data;
			break;

		case N2K_FUNCTION_TRANSDUCER:

			if (!valuesTankInit(system))
				goto error;

			veStrNewFormat(&s, "com.victronenergy.tank.%s_di%d_uc%d", buf,
							N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high),
							J1939_ACL_L_IDENTITY_NUMBER(dev->nmt.name.low));
			system->id = s.data;
			break;

		default:
			goto error;
		}
	}

	sysItt = system;
	while (sysItt) {
		veVariantStr(&sysItt->processName.variant, pltProgramName());
		veVariantStr(&sysItt->processVersion.variant, pltProgramVersion());
		veVariantStr(&sysItt->connection.variant, "VE.Can");
		valuesAddItem(&sysItt->values, "Mgmt/ProcessName", &sysItt->processName, &none, 0);
		valuesAddItem(&sysItt->values, "Mgmt/ProcessVersion", &sysItt->processVersion, &none, 0);
		valuesAddItem(&sysItt->values, "Mgmt/Connection", &sysItt->connection, &none, 0);

		valuesAddItem(&sysItt->values, "ProductId", &sysItt->product.productId, &none, 0);
		veItemSetFmt(&sysItt->product.productId, pidFmt, NULL);

		/*
		 * VE.Bus allready has a "ProductName" item
		 */
		sysItt->product.ptrProductName = veItemByUid(&sysItt->values, "ProductName");
		if (!sysItt->product.ptrProductName) {
			sysItt->product.ptrProductName = &sysItt->product.productName;
			valuesAddItem(&sysItt->values, "ProductName", sysItt->product.ptrProductName, &none, 0);
		}

		valuesAddItem(&sysItt->values, "FirmwareVersion", &sysItt->product.firmwareVersion, &none, 0);
		valuesAddItem(&sysItt->values, "HardwareVersion", &sysItt->product.hardwareVersion, &none, 0);

		/* Victron version formatter */
		if (J1939_ACL_L_GET_MAN(dev->nmt.name.low) == N2K_MAN_VICTRON) {
			veItemSetFmt(&sysItt->product.firmwareVersion, versionFmt, NULL);
			veItemSetFmt(&sysItt->product.hardwareVersion, versionFmt, NULL);
		}

		veVariantUn8(&sysItt->product.instance.variant, N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high));
		valuesAddItem(&sysItt->values, "DeviceInstance", &sysItt->product.instance, &none, 0);
		veItemSetSetter(&sysItt->product.instance, veCanSetDeviceInstanceMasterOfSytemRequest, sysItt);
		veVariantUn8(&sysItt->product.connected.variant, 1);
		valuesAddItem(&sysItt->values, "Connected", &sysItt->product.connected, &none, 0);
		valuesAddItem(&sysItt->values, "Serial", &sysItt->product.serial, &none, 0);
		veVariantUn32(&sysItt->product.n2kUnique.variant, J1939_ACL_L_GET_IDENTITY_NUMBER(dev->nmt.name.low));
		valuesAddItem(&sysItt->values, "N2kUniqueNumber", &sysItt->product.n2kUnique, &none, 0);

		veAssert(sysItt->id != NULL);

		logI(MODULE, "found %s", sysItt->id);
		sysItt->function = fields->function;

		veItemAddChildByUid(itemTree, sysItt->id, &sysItt->values);
		if (!exportToDbus(sysItt->id, &sysItt->values)) {
			logE(MODULE, "export to dbus failed -> exit");
			exit(2);
		}

		/* Link child items, so we can access them if the parent item is updated.
		 * This sets the item context. It does not work with vebus, as vebus items
		 * also use the item context. */
		if (sysItt->child) {
			setChildCtx(product.firmwareVersion);
			setChildCtx(product.hardwareVersion);
			setChildCtx(product.productId);
			setChildCtx(product.serial);
			setChildCtx(product.instance);
			setChildCtx(product.productName);
			setChildCtx(product.connected);
		}

		/* Next sub system */
		sysItt = sysItt->child;
	}

	/* Queue the parent system */
	system->next = systems;
	systems = system;

	return;

error:
	logI(MODULE, "error %s", system->id);
	if (system->child)
		free(system->child);
	free(system);
}


void veCanNewDeviceCallback(struct J1939Func *dev)
{
	VeStr id;
	VeItem *deviceItem;
	VeItem *instanceItem;
	VeVariant v;

	veStrNewFormat(&id, "%08X", dev->nmt.name.low);
	deviceItem = veItemGetOrCreateUid(devicesItem, veStrCStr(&id));
	veStrFree(&id);

	instanceItem = veItemCreateBasic(deviceItem, "Instance", veVariantUn8(&v, N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high)));
	veItemSetSetter(instanceItem, veCanSetDeviceInstanceRequest, dev);
	veItemAddBasic(deviceItem, "ModelName", &dev->n2kProductName);

	veCanCheckForNewDevice(dev);
}

void veCanInstanceChangedCallback(struct J1939Func *dev)
{
	VeStr id;
	VeVariant v;
	VeItem *item;

	veStrNewFormat(&id, "%08X/Instance", dev->nmt.name.low);
	item = veItemByUid(devicesItem, veStrCStr(&id));
	if (item == NULL)
		logE("values", "ignoring instance change of non-existing instance");
	else
		veItemOwnerSet(item, veVariantUn8(&v, N2K_ACL_H_GET_DEVICE_INST(dev->nmt.name.high)));
	veStrFree(&id);
}

/*
 * Returns number of devices in the same group and instance same as the system master device.
 */
un32 veCanCountDevices(VeSystem *system)
{
	un32 count = 0;
	J1939FuncTp *itt;

	itt = j1939FuncList;
	while (itt)
	{
		if (j1939NmtIsOnBus(itt) && (itt->udf.group == system->master->udf.group) &&
				(N2K_ACL_H_GET_DEVICE_INST(itt->nmt.name.high) == system->instance))
			count++;
		itt = itt->next;
	}

	return count;
}

/* Count the number of masters in a system */
un32 veCanCountMasters(VeSystem *system)
{
	un32 count = 0;
	J1939FuncTp *itt;

	itt = j1939FuncList;
	while (itt)
	{
		if (j1939NmtIsOnBus(itt) && itt->isMaster && (itt->udf.group == system->master->udf.group) &&
				(N2K_ACL_H_GET_DEVICE_INST(itt->nmt.name.high) == system->instance))
			count++;
		itt = itt->next;
	}

	return count;
}

static void freeSystem(VeSystem *system)
{
	logI(MODULE, "Removing %s", system->id);
	logE(MODULE, "Removing dbus services currently not supported: exit");
	exit(1);

#if 0
	veDbusDisconnect(system->values.dbus);

	if (system->canbusCtx)
		free(system->canbusCtx);
	free(system);
#endif
}

VeSystem *veCanRemoveSystem(VeSystem *system)
{
	VeSystem *s = systems;

	if (system == systems) {
		systems = system->next;
		freeSystem(system);
		return systems;
	}

	while (s) {
		if (s->next == system) {
			s->next = system->next;
			freeSystem(system);
			return s->next;
		}
		s = s->next;
	}

	return s;
}
