#include <stdlib.h>

#include <velib/base/base.h>
#include <velib/base/ve_string.h>
#include <velib/types/ve_item.h>

#include <dbus_service.h>
#include <values_lynxion.h>
#include <values_common.h>
#include <values_battery.h>
#include <vecan_lynxion.h>
#include <vecan_battery.h>

/**
 * Definitions how to make a textual representation of a value.
 * @see veVariantFmt in types/ve_variant_print.c.
 */
static VeVariantUnitFmt dcV = {2, "V"};
static VeVariantUnitFmt dcI = {1, "A"};
static VeVariantUnitFmt dcW = {0, "W"};
static VeVariantUnitFmt dcT = {0, " degrees"};
static VeVariantUnitFmt ampH = {1, "Ah"};
static VeVariantUnitFmt kwh = {2, "kWh"};
static VeVariantUnitFmt sec = {0, "s"};
static VeVariantUnitFmt none = {0, ""};
static VeVariantUnitFmt percent = {1, "%"};

#define offset(a)	offsetof(VeLynxIon, a)
static ItemInfo itemInfoBms[] = {
	{"State",											&none,	offset(state),					8	},
	{"Error",											&none,	offset(error),					8	},
	{"SystemSwitch",									&none,	offset(systemSwitch),			8	},
	{"Balancing",										&none,	offset(balancing),				8	},
	{"System/NrOfBatteries",							&none,	offset(nrOfBatteries),			8	},
	{"System/BatteriesParallel",						&none,	offset(batteriesParallel),		8	},
	{"System/BatteriesSeries",							&none,	offset(batteriesSeries),		8	},
	{"System/NrOfCellsPerBattery",						&none,	offset(nrOfCellsPerBattery),	8	},
	{"System/MinCellVoltage",							&dcV,	offset(minCellVoltage),			8	},
	{"System/MaxCellVoltage",							&dcV,	offset(maxCellVoltage),			8	},
	{"Diagnostics/ShutDownsDueError",					&none,	offset(shutdownsDueError),		8	},
	{"Diagnostics/LastError1",							&none,	offset(lastError1),				8	},
	{"Diagnostics/LastError2",							&none,	offset(lastError2),				8	},
	{"Diagnostics/LastError3",							&none,	offset(lastError3),				8	},
	{"Diagnostics/LastError4",							&none,	offset(lastError4),				8	},
	{"Io/AllowToCharge",								&none,	offset(allowToCharge),			8	},
	{"Io/AllowToDischarge",								&none,	offset(allowToDischarge),		8	},
	{"Io/ExternalRelay",								&none,	offset(externalRelay),			8	},
	{"History/MinimumCellVoltage",						&dcV,	offset(history.minCellVoltage),	8	},
	{"History/MaximumCellVoltage",						&dcV,	offset(history.maxCellVoltage),	8	},
};
#undef offset

#define offset(a)	offsetof(VeBattery, a)
static ItemInfo itemInfoBat[] = {
	{"Dc/0/Voltage",									&dcV,	offset(dcStatusBattery.voltage),	8	},
	{"Dc/0/Current",									&dcI,	offset(dcStatusBattery.current),	8	},
	{"Dc/0/Power",										&dcW,	offset(dcStatusBattery.power),		8	},
	{"Dc/0/Temperature",								&dcT,	offset(batteryTemperature),			8	},
	{"ConsumedAmphours",								&ampH,	offset(consumedAmphours),			8	},
	{"Soc",												&percent,	offset(stateOfCharge),			8	},
	{"TimeToGo",										&sec,	offset(timeToGo),					8	},
	{"History/DeepestDischarge",						&ampH,	offset(history.h[0]),				8	},
	{"History/LastDischarge",							&ampH,	offset(history.h[1]),				8	},
	{"History/AverageDischarge",						&ampH,	offset(history.h[2]),				8	},
	{"History/ChargeCycles",							&none,	offset(history.h[3]),				8	},
	{"History/FullDischarges",							&none,	offset(history.h[4]),				8	},
	{"History/TotalAhDrawn",							&ampH,	offset(history.h[5]),				8	},
	{"History/MinimumVoltage",							&dcV,	offset(history.h[6]),				8	},
	{"History/MaximumVoltage",							&dcV,	offset(history.h[7]),				8	},
	{"History/TimeSinceLastFullCharge",					&sec,	offset(history.h[8]),				8	},
	{"History/AutomaticSyncs",							&none,	offset(history.h[9]),				8	},
	{"History/DischargedEnergy",						&kwh,	offset(history.h[16]),				8	},
	{"History/ChargedEnergy",							&kwh,	offset(history.h[17]),				8	},
//	{"History/Clear",									&none,	offset(history.clear),				0	},
	{"Settings/HasTemperature",							&none,	offset(hasTemperature),				8	},
	{"Settings/HasSettings",							&none,	offset(hasSettings),				8	},
	{"Settings/DischargeFloorLinkedToRelay",			&none,	offset(dischargeFloorLinkedToRelay),8	},

};
#undef offset

/*
 * Create a Lynx Ion system object with the corresponding item tree.
 *
 * The tree with items can be used to navigate through the values,
 * format / set / get them etc.
 */
veBool valuesLynxIonInit(VeSystem *system)
{
	VeLynxIon *lynxIon = calloc(1, sizeof(VeLynxIon));

	if (!lynxIon)
		return veFalse;

	/* create the values tree to store the variables */
	system->canbusCtx = lynxIon;
	valuesInitTree(itemInfoBms, ARRAY_LENGTH(itemInfoBms), &system->values, system);

	/* setup the CAN bus to fill them in */
	if (!veCanLynxIonAdd(system)) {
		free(lynxIon);
		return veFalse;
	}

	return veTrue;
}

/*
 * The Lynx Ion + Shunt also has a monitor function.
 * So it is connected to the DBUS as two different services.
 */
veBool valuesLynxIonBatteryInit(VeSystem *system)
{
	VeBattery *lynxIon = calloc(1, sizeof(VeBattery));

	if (!lynxIon)
		return veFalse;

	veVariantUn32(&lynxIon->hasTemperature.variant, 1);
	veVariantUn32(&lynxIon->hasSettings.variant, 0);

	/* No settings support */
	lynxIon->minSettingsVersion = 0xFFFF;

	/* Not applicable to Lynx Ion */
	lynxIon->minRelayLowSocVersion = 0;

	/* create the values tree to store the variables */
	system->canbusCtx = lynxIon;
	valuesInitTree(itemInfoBat, ARRAY_LENGTH(itemInfoBat), &system->values, system);

	/* setup the CAN bus to fill them in */
	if (!veCanBatteryAdd(system)) {
		free(lynxIon);
		return veFalse;
	}

	return veTrue;
}
