#include <stdlib.h>

#include <velib/base/base.h>
#include <velib/base/ve_string.h>
#include <velib/types/ve_item.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/velib_inc_J1939_app.h>

#include <dbus_service.h>
#include <values_battery.h>
#include <values_common.h>
#include <vecan_battery.h>
#include <vecan_common.h>

/**
 * Definitions how to make a textual representation of a value.
 * @see veVariantFmt in types/ve_variant_print.c.
 */
static VeVariantUnitFmt dcV = {2, "V"};
static VeVariantUnitFmt dcI = {1, "A"};
static VeVariantUnitFmt dcW = {0, "W"};
static VeVariantUnitFmt dcT = {0, " degrees"};
static VeVariantUnitFmt kelvin = {0, "K"};
static VeVariantUnitFmt kwh = {2, "kWh"};
static VeVariantUnitFmt ampH = {1, "Ah"};
static VeVariantUnitFmt sec = {0, "s"};
static VeVariantUnitFmt min = {0, "min"};
static VeVariantUnitFmt none = {0, ""};
static VeVariantUnitFmt pct = {1, "%"};

/* Scales used for settings */
#define unitary NULL
static VeScale const div10 = { 0, 0.1 };
static VeScale const div100 = { 0, 0.01 };

/* Ranges used for settings */

static VeRange const capacity = { 1, 9999, 200 };
static VeRange const cef = { 50, 99, 95 };
static VeRange const currThd = { 0, 2, 0.1 };
static VeRange const fuse = { 0, 1, 1 };
static VeRange const hTemp = { 0, 373, 45 + 273.15 };
static VeRange const hTempClr = { 0, 373, 30 + 273.15};
static VeRange const lSoc = { 0, 99, 50 };
static VeRange const lSocClr = { 0, 99, 90 };
static VeRange const lTemp = { 0, 373, -20 + 273.15 };
static VeRange const lTempClr = { 0, 373, -10 + 273.15 };
static VeRange const minutes = { 0, 500, 0 };
static VeRange const minutes2 = { 1, 50, 3 };
static VeRange const peukert = { 1.00, 1.50, 1.25 };
static VeRange const relayMode = { 0, 3, 0 };
static VeRange const relayState = { 0, 1, 0 };
static VeRange const tailCurr = { 0.5, 10.0, 4.0 };
static VeRange const ttgDT = { 0, 12, 3 };
static VeRange const voltage = { 0, 90, 0 };
static VeRange const zero = { 0, 0, 0 };

/**
 * The tree of known values of a BMV.
 */

#define offset(a)	offsetof(VeBattery, a)
static ItemInfo itemInfo[] = {
	{"Dc/0/Voltage",								&dcV,	offset(dcStatusBattery.voltage),		8	},
	{"Dc/0/Current",								&dcI,	offset(dcStatusBattery.current),		8	},
	{"Dc/0/Power",									&dcW,	offset(dcStatusBattery.power),			8	},
	{"Dc/0/Temperature",							&dcT,	offset(batteryTemperature),				8	},
	{"Dc/1/Voltage",								&dcV,	offset(dcStatusAux.voltage),			8	},
	{"ConsumedAmphours",							&ampH,	offset(consumedAmphours),				8	},
	{"Soc",											&pct,	offset(stateOfCharge),					8	},
	{"TimeToGo",									&sec,	offset(timeToGo),						8	},
	{"Alarms/Alarm",								&none,	offset(alarms.alarm),					8	},
	{"Alarms/LowVoltage",							&none,	offset(alarms.lowVoltage),				8	},
	{"Alarms/HighVoltage",							&none,	offset(alarms.highVoltage),				8	},
	{"Alarms/LowSoc",								&none,	offset(alarms.lowSoc),					8	},
	{"Alarms/FuseBlown",							&none,	offset(alarms.fuseBlown),				8	},
	{"Alarms/HighTemperature",						&none,	offset(alarms.highBatteryTemperature),	8	},
	{"Alarms/LowTemperature",						&none,	offset(alarms.lowBatteryTemperature),	8	},
	{"Alarms/HighInternalTemperature",				&none,	offset(alarms.highInternalTemperature),	8	},
	{"History/DeepestDischarge",					&ampH,	offset(history.h[0]),					8	},
	{"History/LastDischarge",						&ampH,	offset(history.h[1]),					8	},
	{"History/AverageDischarge",					&ampH,	offset(history.h[2]),					8	},
	{"History/ChargeCycles",						&none,	offset(history.h[3]),					8	},
	{"History/FullDischarges",						&none,	offset(history.h[4]),					8	},
	{"History/TotalAhDrawn",						&ampH,	offset(history.h[5]),					8	},
	{"History/MinimumVoltage",						&dcV,	offset(history.h[6]),					8	},
	{"History/MaximumVoltage",						&dcV,	offset(history.h[7]),					8	},
	{"History/TimeSinceLastFullCharge",				&sec,	offset(history.h[8]),					8	},
	{"History/AutomaticSyncs",						&none,	offset(history.h[9]),					8	},
	{"History/LowVoltageAlarms",					&none,	offset(history.h[10]),					8	},
	{"History/HighVoltageAlarms",					&none,	offset(history.h[11]),					8	},
	{"History/DischargedEnergy",					&kwh,	offset(history.h[16]),					8	},
	{"History/ChargedEnergy",						&kwh,	offset(history.h[17]),					8	},
	{"History/Clear",								&none,	offset(history.clear),					0	},

	{"Settings/HasTemperature",						&none,	offset(hasTemperature),					0	},
	{"Settings/HasSettings",						&none,	offset(hasSettings),					8	},
	{"Settings/DischargeFloorLinkedToRelay",		&none,	offset(dischargeFloorLinkedToRelay),	8	},
	{"Settings/RestoreDefaults",					&none,	offset(restoreDefaults),				0	},

#undef offset
#define offset(a)	offsetof(VeBattery, settings.a)

	{"Settings/Battery/Locked",						&none,	offset(battery.locked),					8,	},
	{"Settings/Battery/Capacity",					&ampH,	offset(battery.capacity),				0,	},
	{"Settings/Battery/ChargedVoltage",				&dcV,	offset(battery.chargedVoltage),			0,	},
	{"Settings/Battery/TailCurrent",				&pct,	offset(battery.tailCurrent),			0,	},
	{"Settings/Battery/ChargedDetectionTime",		&min,	offset(battery.chargedDetectionTime),	0,	},
	{"Settings/Battery/ChargeEfficiency",			&pct,	offset(battery.chargeEfficiency),		0,	},
	{"Settings/Battery/PeukertExponent",			&none,	offset(battery.peukertExponent),		0,	},
	{"Settings/Battery/CurrentThreshold",			&dcI,	offset(battery.currentThreshold),		0,	},
	{"Settings/Battery/TTGAveragingPeriod",			&min,	offset(battery.ttgDeltaT),				0,	},
	{"Settings/Battery/DischargeFloor",				&pct,	offset(battery.dischargeFloor),			0,	},
	{"Settings/Battery/CurrentOffset",				&none,	offset(battery.currentOffset),			0,	},
	{"Settings/Battery/Synchronize",				&none,	offset(battery.sync),					0	},
	{"Settings/Battery/ZeroCurrent",				&none,	offset(battery.zero),					0	},

	{"Relay/0/State",								&none,	offset(relay.state),					8,	},
	{"Settings/Relay/Mode",							&none,	offset(relay.mode),						0,	},
	{"Settings/Relay/MinCloseTime",					&min,	offset(relay.minEnabledTime),			0,	},
	{"Settings/Relay/OffDelay",						&min,	offset(relay.offDelay),					0,	},
	{"Settings/Relay/LowSoc",						&pct,	offset(relay.lowSoc),					0,	},
	{"Settings/Relay/LowSocClear",					&pct,	offset(relay.lowSocClear),				0,	},
	{"Settings/Relay/LowVoltage",					&dcV,	offset(relay.lowVoltage),				0,	},
	{"Settings/Relay/LowVoltageClear",				&dcV,	offset(relay.lowVoltageClear),			0,	},
	{"Settings/Relay/HighVoltage",					&dcV,	offset(relay.highVoltage),				0,	},
	{"Settings/Relay/HighVoltageClear",				&dcV,	offset(relay.highVoltageClear),			0,	},
	{"Settings/Relay/FuseBlown",					&none,	offset(relay.fuseBlown),				0,	},
	{"Settings/Relay/HighBatteryTemperature",		&kelvin,offset(relay.highBatteryTemp),			0,	},
	{"Settings/Relay/HighBatteryTemperatureClear",	&kelvin,offset(relay.highBatteryTempClear),		0,	},
	{"Settings/Relay/LowBatteryTemperature",		&kelvin,offset(relay.lowBatteryTemp),			0,	},
	{"Settings/Relay/LowBatteryTemperatureClear",	&kelvin,offset(relay.lowBatteryTempClear),		0,	},

	{"Settings/Alarm/LowSoc",						&pct,	offset(alarm.lowSoc),					0,	},
	{"Settings/Alarm/LowSocClear",					&pct,	offset(alarm.lowSocClear),				0,	},
	{"Settings/Alarm/LowVoltage",					&dcV,	offset(alarm.lowVoltage),				0,	},
	{"Settings/Alarm/LowVoltageClear",				&dcV,	offset(alarm.lowVoltageClear),			0,	},
	{"Settings/Alarm/HighVoltage",					&dcV,	offset(alarm.highVoltage),				0,	},
	{"Settings/Alarm/HighVoltageClear",				&dcV,	offset(alarm.highVoltageClear),			0,	},
	{"Settings/Alarm/HighBatteryTemperature",		&kelvin,offset(alarm.highBatteryTemp),			0,	},
	{"Settings/Alarm/HighBatteryTemperatureClear",	&kelvin,offset(alarm.highBatteryTempClear),		0,	},
	{"Settings/Alarm/LowBatteryTemperature",		&kelvin,offset(alarm.lowBatteryTemp),			0,	},
	{"Settings/Alarm/LowBatteryTemperatureClear",	&kelvin,offset(alarm.lowBatteryTempClear),		0,	},
};

static veBool locked(void *ctx)
{
	VeBattery *battery = (VeBattery *)ctx;
	VeVariant locked;

	veItemValue(&battery->settings.battery.locked, &locked);
	return veVariantIsValid(&locked) && locked.value.UN32 == 1;
}

/* Settings definition: vreg, type, min, max, scale */
static VeSetting settingInfo[] = {
	{VE_REG_BATTERY_CAPACITY,			offset(battery.capacity),				VE_UN16,	&capacity,	unitary,	&locked	},
	{VE_REG_CHARGED_VOLTAGE,			offset(battery.chargedVoltage),			VE_UN16,	&voltage,	&div10,		&locked	},
	{VE_REG_CHARGED_CURRENT,			offset(battery.tailCurrent),			VE_UN16,	&tailCurr,	&div10,		&locked	},
	{VE_REG_CHARGE_DETECTION_TIME,		offset(battery.chargedDetectionTime),	VE_UN16,	&minutes2,	unitary,	&locked	},
	{VE_REG_CHARGE_EFFICIENCY,			offset(battery.chargeEfficiency),		VE_UN16,	&cef,		unitary,	&locked	},
	{VE_REG_PEUKERT_COEFFICIENT,		offset(battery.peukertExponent),		VE_UN16,	&peukert,	&div100,	&locked	},
	{VE_REG_CURRENT_THRESHOLD,			offset(battery.currentThreshold),		VE_UN16,	&currThd,	&div100,	&locked	},
	{VE_REG_TTG_DELTA_T,				offset(battery.ttgDeltaT),				VE_UN16,	&ttgDT,		unitary,	&locked	},
	{VE_REG_LOW_SOC,					offset(battery.dischargeFloor),			VE_UN16,	&lSoc,		&div10,		&locked	},
	{VE_REG_LOW_SOC_CLEAR,				offset(battery.lowSocClear),			VE_UN16,	&lSocClr,	&div10,		&locked	},
	{VE_REG_CURRENT_OFFSET,				offset(battery.currentOffset),			VE_SN16,	&zero,		unitary,	&locked	},
	{VE_REG_SYNCHRONIZE,				offset(battery.sync),					VE_UN8,		&zero,		unitary,	&locked	},
	{VE_REG_ZERO_CURRENT,				offset(battery.zero),					VE_UN8,		&zero,		unitary,	0		},

	{VE_REG_RELAY_MODE,					offset(relay.mode),						VE_UN16,	&relayMode,	unitary,	0		},
	{VE_REG_RELAY_CONTROL,				offset(relay.state),					VE_UN8,		&relayState,unitary,	0		},
	{VE_REG_RELAY_MIN_ENABLED,			offset(relay.minEnabledTime),			VE_UN16,	&minutes,	unitary,	0		},
	{VE_REG_RELAY_DISABLE_DELAY,		offset(relay.offDelay),					VE_UN16,	&minutes,	unitary,	0		},
	{VE_REG_RELAY_LOW_SOC_SET,			offset(relay.lowSoc),					VE_UN16,	&lSoc,		&div10,		0		},
	{VE_REG_RELAY_LOW_SOC_CLEAR,		offset(relay.lowSocClear),				VE_UN16,	&lSocClr,	&div10,		0		},
	{VE_REG_RELAY_LOW_VOLTAGE_SET,		offset(relay.lowVoltage),				VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_LOW_VOLTAGE_CLEAR,	offset(relay.lowVoltageClear),			VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_HIGH_VOLTAGE_SET,		offset(relay.highVoltage),				VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_HIGH_VOLTAGE_CLEAR,	offset(relay.highVoltageClear),			VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_LOW_VOLTAGE_2_SET,	offset(relay.lowStarterVoltage),		VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_LOW_VOLTAGE_2_CLEAR,	offset(relay.lowStarterVoltageClear),	VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_HIGH_VOLTAGE_2_SET,	offset(relay.highStarterVoltage),		VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_HIGH_VOLTAGE_2_CLEAR,	offset(relay.highStarterVoltageClear),	VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_RELAY_FUSE_BLOWN,			offset(relay.fuseBlown),				VE_UN16,	&fuse,		unitary,	0		},
	{VE_REG_RELAY_HIGH_BAT_TEMP_SET,	offset(relay.highBatteryTemp),			VE_UN16,	&hTemp,		&div100,	0		},
	{VE_REG_RELAY_HIGH_BAT_TEMP_CLEAR,	offset(relay.highBatteryTempClear),		VE_UN16,	&hTempClr,	&div100,	0		},
	{VE_REG_RELAY_LOW_BAT_TEMP_SET,		offset(relay.lowBatteryTemp),			VE_UN16,	&lTemp,		&div100,	0		},
	{VE_REG_RELAY_LOW_BAT_TEMP_CLEAR,	offset(relay.lowBatteryTempClear),		VE_UN16,	&lTempClr,	&div100,	0		},

	{VE_REG_ALARM_LOW_SOC_SET,			offset(alarm.lowSoc),					VE_UN16,	&lSoc,		&div10,		0		},
	{VE_REG_ALARM_LOW_SOC_CLEAR,		offset(alarm.lowSocClear),				VE_UN16,	&lSocClr,	&div10,		0		},
	{VE_REG_ALARM_LOW_VOLTAGE_SET,		offset(alarm.lowVoltage),				VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_LOW_VOLTAGE_CLEAR,	offset(alarm.lowVoltageClear),			VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_HIGH_VOLTAGE_SET,		offset(alarm.highVoltage),				VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_HIGH_VOLTAGE_CLEAR,	offset(alarm.highVoltageClear),			VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_LOW_VOLTAGE_2_SET,	offset(alarm.lowStarterVoltage),		VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_LOW_VOLTAGE_2_CLEAR,	offset(alarm.lowStarterVoltageClear),	VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_HIGH_VOLTAGE_2_SET,	offset(alarm.highStarterVoltage),		VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_HIGH_VOLTAGE_2_CLEAR,	offset(alarm.highStarterVoltageClear),	VE_UN16,	&voltage,	&div100,	0		},
	{VE_REG_ALARM_HIGH_BAT_TEMP_SET,	offset(alarm.highBatteryTemp),			VE_UN16,	&hTemp,		&div100,	0		},
	{VE_REG_ALARM_HIGH_BAT_TEMP_CLEAR,	offset(alarm.highBatteryTempClear),		VE_UN16,	&hTempClr,	&div100,	0		},
	{VE_REG_ALARM_LOW_BAT_TEMP_SET,		offset(alarm.lowBatteryTemp),			VE_UN16,	&lTemp,		&div100,	0		},
	{VE_REG_ALARM_LOW_BAT_TEMP_CLEAR,	offset(alarm.lowBatteryTempClear),		VE_UN16,	&lTempClr,	&div100,	0		},

	{VE_REG_CLEAR_HISTORY,				offsetof(VeBattery, history.clear),		VE_UN8,		&zero,		unitary,	0		},
	{VE_REG_INVALID}
};

#undef offset

static veBool readOnly(VeItem *item, void *ctx, VeVariant *variant)
{
	VE_UNUSED(item);
	VE_UNUSED(ctx);
	VE_UNUSED(variant);

	return veFalse;
}

static veBool setRelayLowSocSet(VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem *)ctx;
	VeBattery *battery = (VeBattery *)system->canbusCtx;

	/*
	 * Older software versions (< v1.09) do not have a separate relay threshold.
	 * The discharge floor is used for this purpose because it is linked to the relay.
	 */
	VeItem *dischargeFloorLinkedToRelay = &battery->dischargeFloorLinkedToRelay;
	if (veItemIsValid(dischargeFloorLinkedToRelay) && dischargeFloorLinkedToRelay->variant.value.UN32)
		item = &battery->settings.battery.dischargeFloor;

	VeSetting *setting;
	if ((setting = valuesGetSettingByItem(&system->settingsCtx, item)))
		return veCanSetVeReg(system, variant, setting);

	return veFalse;
}

static veBool setRelayLowSocClear(VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem *)ctx;
	VeBattery *battery = (VeBattery *)system->canbusCtx;

	/* Older software versions (< v1.09) do not have a separate relay threshold. */
	VeItem *dischargeFloorLinkedToRelay = &battery->dischargeFloorLinkedToRelay;
	if (veItemIsValid(dischargeFloorLinkedToRelay) && dischargeFloorLinkedToRelay->variant.value.UN32)
		item = &battery->settings.battery.lowSocClear;

	VeSetting *setting;
	if ((setting = valuesGetSettingByItem(&system->settingsCtx, item)))
		return veCanSetVeReg(system, variant, setting);

	return veFalse;
}

static veBool setAlarmLowSocThreshold(VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem *)ctx;
	VeBattery *battery = (VeBattery *)system->canbusCtx;

	/* Lynx Shunt <= v1.08 has bug in the low Soc alarm threshold: it is off by a factor 100 */
	if (battery->hasLowSocAlarmBug) {
		veVariantToN32(variant);

		/* Clamp at 90, as max is 99%. This would be rounded to 100%,
		 * but is not supported. Add 5% for rounding to nearest integer */
		if (variant->value.UN32 > 90)
			variant->value.UN32 = 90;
		variant->value.UN32 += 5;

		/* Correct the factor 100. E.g. 50% would become 0.5%. veCanSetVeReg divides it by 0.1
		 * for the scale on the canbus. So 0.5 / 0.1 = 5. Lynx Shunt: 5 * 0.1 * 100 = 50% */
		veVariantToFloat(variant);
		variant->value.Float *= 0.01;
	}

	VeSetting *setting;
	if ((setting = valuesGetSettingByItem(&system->settingsCtx, item)))
		return veCanSetVeReg(system, variant, setting);

	return veFalse;
}

static veBool setRestoreDefaults(VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem *)ctx;

	veVariantToN32(variant);
	if (!veVariantIsNumeric(variant) || !veVariantIsValid(variant))
		return veFalse;

	/* Only restore defaults if not equal to 0 */
	if (variant->value.UN32 == 0)
		return veFalse;

	if (!veRegPrepareSf(system->master->nmt.nad, VE_REG_RESTORE_DEFAULT))
		return veFalse;

	veOutSendSf();

	/* Trigger on-change event. This is needed for the gui */
	veItemOwnerSet(item, variant);
	veItemOwnerSet(item, veVariantUn32(variant, 0));

	return veTrue;
}

/*
 * Create a Lynx-shunt system object with the corresponding item tree.
 *
 * The tree with items can be used to navigate through the values,
 * format / set / get them etc.
 */
veBool valuesLynxShuntInit(VeSystem *system)
{
	VeBattery *battery = calloc(1, sizeof(VeBattery));

	if (!battery)
		return veFalse;

	veVariantUn32(&battery->hasTemperature.variant, 1);

	/* By default the battery settings are unlocked.They should be
	 * locked if a Lynx Ion is found as this manages the settings. */
	veVariantUn32(&battery->settings.battery.locked.variant, 0);

	/* And make it read-only */
	veItemSetSetter(&battery->settings.battery.locked, readOnly, NULL);

	/* Support al versions */
	battery->minSettingsVersion = 0;

	/* In the Lynx Shunt VE.Can the relay and discharge floor are separated since v1.09 */
	battery->minRelayLowSocVersion = 0x109;

	/* Settings support is indicated by this item.
	 * Off by default, enabled if > minSettingsVersion. */
	veVariantUn32(&battery->hasSettings.variant, 0);

	/* create the values tree to store the variables */
	system->canbusCtx = battery;
	system->settingsCtx.system = system;
	system->settingsCtx.settings = settingInfo;
	battery->settingsCtx = settingInfo;
	valuesInitTree(itemInfo, ARRAY_LENGTH(itemInfo), &system->values, system);

	/* Override setters */
	veItemSetSetter(&battery->settings.relay.lowSoc, setRelayLowSocSet, system);
	veItemSetSetter(&battery->settings.relay.lowSocClear, setRelayLowSocClear, system);
	veItemSetSetter(&battery->settings.alarm.lowSoc, setAlarmLowSocThreshold, system);
	veItemSetSetter(&battery->settings.alarm.lowSocClear, setAlarmLowSocThreshold, system);

	veVariantUn32(&battery->restoreDefaults.variant, 0); /* Type needed */
	veItemSetSetter(&battery->restoreDefaults, setRestoreDefaults, system);

	/* setup the CAN bus to fill them in */
	if (!veCanBatteryAdd(system)) {
		free(battery);
		return veFalse;
	}

	return veTrue;
}
