#include <stdlib.h>

#include <velib/velib_inc_J1939.h>

#include <velib/base/base.h>
#include <velib/base/ve_string.h>
#include <velib/mk2/mk2_strings.h>
#include <velib/mk2/values_float.h>
#include <velib/types/ve_str.h>
#include <velib/types/ve_item.h>
#include <velib/types/variant_print.h>
#include <velib/utils/ve_logger.h>

#include <values_vebus.h>
#include <values_common.h>
#include <vecan_vebus.h>

static VeVariantUnitFmt none = {0, ""};

/* stub to forward current limit requests for the active input */
static veBool VeItemSetCurrentLimit(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem*) ctx;
	VeBusSys *vebus = (VeBusSys*) system->canbusCtx;
	un16 limit;
	VE_UNUSED(item);

	veVariantToFloat(variant);
	limit = (un16) variant->value.Float * 10;

	if (vebus->items.VE_SYS_AC_IN_ACTIVE.variant.value.UN8 == 0)
		return veCanVebusSetAcLimitAcIn1(system, limit);
	if (vebus->items.VE_SYS_AC_IN_ACTIVE.variant.value.UN8 == 1)
		return veCanVebusSetAcLimitAcIn2(system, limit);

	return veFalse;
}

/* stub to forward state change request. */
static veBool VeItemSetMode(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem*) ctx;
	VE_UNUSED(item);

	veCanVebusSetMode(system, variant->value.SN32);

	return veTrue;
}

/* stub to forward max charge current changes */
static veBool VeItemSetMaxChargeCurrent(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VeSystem *system = (VeSystem *) ctx;
	un16 limit;
	VE_UNUSED(item);

	veVariantToFloat(variant);
	limit = (un16) variant->value.Float * 10;

	veCanVebusSetMaxChargeCurrent(system, limit);

	return veTrue;
}

/* textual representation of a vebus value */
static size_t textVersion(VeVariant *var, void const *ctx, char *buf, size_t len)
{
	VeItem *item = (VeItem *) ctx;
	un16 nr = item->ctx.value;

	if (veVariantIsValid(var)) {
		VebusUnit unit = vebusGetUnit(nr, VE_UNITS_VSC);

		/* The dbus uses percentage as 0..100 not 0..1 */
		if (unit == VE_U_PERC_FRAC)
			unit = VE_U_PERC;

		return vebusFormatValue(unit, var, buf, len);
	}

	return ve_snprintf(buf, len, "%s", "");
}

size_t getDescription(struct VeItem *item, char *buf, size_t len)
{
	return ve_snprintf(buf, len, "%s", vebusGetName(item->ctx.value));
}

/* create a item corresponding to a vebus value */
static void addItem(struct VeItem *root, un16 nr, char const *uid, struct VeItem *item)
{
	veItemAddChildByUid(root, uid, item);
	veItemSetFmt(item, textVersion, item);
	veItemSetGetDescr(item, getDescription);
	item->variant.type.tp = VE_FLOAT;
	veItemInvalidate(item);

	item->ctx.value = nr;
}

/*
 * Create a vebus system object and the corresponding item tree.
 *
 * The tree with items can be used to navigate through the values in
 * vebus system, format / set / get them etc.
 */
veBool valuesVebusInit(VeSystem *system)
{
	VeBusSys *vebus = calloc(1, sizeof(VeBusSys));
	VeVariant v;
	VeItem *item;

	if (vebus == NULL)
		return veFalse;

#define X(a,b)	addItem(&system->values, a, b, &vebus->items.a);
	VE_SYSTEM_CREATE_ITEMS
#undef X

	system->canbusCtx = vebus;

	veItemAddChildByUid(&system->values, "Hub1/ChargeVoltage", &vebus->hub1VoltageSetPoint);

	if (!veCanVebusAdd(system)) {
		free(vebus);
		return veFalse;
	}

	/* Create a dbus mk2 device and listen to the mk2 on the CAN bus */
	veCanRemoteDevInit(&vebus->mk2Can, &system->master->nmt.nad, 0, VE_REMOTE_DEV_ROLE_CLIENT_PROXY);
	item = veItemGetOrCreateUid(&system->values, "Interfaces/Mk2/Tunnel");
	veItemRemoteDevInit(&vebus->mk2Dbus, item, VE_REMOTE_DEV_ROLE_SERVICE_PROXY);
	veRemoteDeviceCreateTunnel(&vebus->mk2Dbus.remoteDev, &vebus->mk2Can.remoteDev);

	veItemSetSetter(&vebus->items.VE_SYS_AC_IN_LIM, VeItemSetCurrentLimit, system);
	veItemSetSetter(&vebus->items.VE_SYS_MODE, VeItemSetMode, system);
	veItemSetSetter(&vebus->items.VE_I_BAT_BULK, VeItemSetMaxChargeCurrent, system);

	veItemSetFmt(&vebus->items.VE_SYS_MK2_VERSION, veVariantFmt, &none);
	veItemSetFmt(&vebus->items.VE_SYS_VEBUS_VERSION, veVariantFmt, &none);

	/* For now assume current limit and mode can always be controlled */
	veItemOwnerSet(&vebus->items.VE_SYS_MODE_IS_ADJUSTABLE, veVariantBit(&v, 2, veTrue));
	veItemOwnerSet(&vebus->items.VE_SYS_CURRENT_LIMIT_IS_ADJUSTABLE, veVariantBit(&v, 2, veTrue));

	return veTrue;
}
